//-----------------------------------------------------------------------------------
// Modified version of LEDA's _dlist.c, dlist.h, list.h.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#ifndef BM_DLIST_H
#define BM_DLIST_H

#include <LEDA/basic.h>

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

class bmdlist; 
class bmdlink; 
typedef bmdlink* bmlist_item;

class bmdlink {

public:
  bmdlink* ptr1;
  bmdlink* ptr2;
  GenPtr e;

  bmdlink(GenPtr a, bmdlink* pre, bmdlink* suc)
  { 
    e = a;
    ptr1 = suc;
    ptr2 = pre;
  }

  LEDA_MEMORY(bmdlink)
  friend class bmdlist;
};

class bmdlist {

public:

  bmdlink* h;                //head
  bmdlink* t;                //tail
  int count;                 //length of list

  bmdlist()      
    {
      h = t = 0;
      count = 0;
    }

  bmdlink* first_item() { return h; }
  bmdlink* last_item() { return t; }
  GenPtr& entry(bmdlink* it) { return it->e; }
  int size() { return count; }

  bmdlink* next_item(bmdlink* p, bmdlink* pred_p)
    {
      if (p)
	{
	  if (p == t)
	    return 0;
	  if (p == h)
	    pred_p = 0;
	  if (p->ptr1 != pred_p)
	    return p->ptr1;
	  else
	    return p->ptr2;
	}
      else return 0;
    }
  bmdlink* pred_item(bmdlink* p, bmdlink* succ_p)
    {
      if (p)
	{
	  if (p == h)
	    return 0;
	  if (p == t)
	    succ_p = 0;
	  if (p->ptr2 != succ_p)
	    return p->ptr2;
	  else
	    return p->ptr1;
	}
      else return 0;
    }
  bmdlink* cyclic_pred(bmdlink* it, bmdlink* succ_it)
    {
      if (it == 0)
	return 0;
      if (it == h)
	return t;
      if (it == t)
	succ_it = 0;
      return pred_item(it, succ_it);
    }
  bmdlink* cyclic_succ(bmdlink* it, bmdlink* pred_it)
    {
      if (it == 0)
	return 0;
      if (it == t)
	return h;
      if (it == h)
	pred_it = 0;
      return next_item(it, pred_it);
    }

  bmdlink* push(GenPtr a)   
    {
      count++;
      if (h)
	{
	  if (h->ptr2)
	    h = h->ptr1 = new bmdlink(a,0,h);
	  else
	    h = h->ptr2 = new bmdlink(a,0,h);
	}
      else
	h = t =  new bmdlink(a,0,0);
      return h;
    }
  bmdlink* append(GenPtr a)
    { 
      count++;
      if (t)
	{
	  if (t->ptr1)
	    t = t->ptr2 = new bmdlink(a,t,0);
	  else
	    t = t->ptr1 = new bmdlink(a,t,0);
	}
      else
	t = h = new bmdlink(a,0,0);
      return t; 
    }
 
  GenPtr del_item(bmdlink* it)
    { 
      if (it == 0)
	error_handler(999,"dlist: delete nil-item");
      if (it == h)
	return pop();
      if (it == t)
	return pop_back();
      bmdlink* p = pred_item(it, 0);
      bmdlink* s = next_item(it, p);
      GenPtr x = it->e;
      if (p->ptr1 == it)
	p->ptr1 = s;
      else
	p->ptr2 = s;
      if (s->ptr2 == it)
	s->ptr2 = p;
      else
	s->ptr1 = p;
      count--;
      delete it;
      return x;
    }
  GenPtr pop()    
    {
      if (h == 0) 
	error_handler(1,"dlist: pop on empty list.");
      bmdlink* x = h;
      h = next_item(h, nil);
      if (h)
	{
	  if (h->ptr2 == x)
	    h->ptr2 = 0;
	  else
	    h->ptr1 = 0;
	}
      else
	t = 0;
      GenPtr p = x->e;
      delete x;
      count--;
      return p;
    }
  GenPtr pop_back()    
    {
      if (h == 0) 
	error_handler(1,"dlist: Pop on empty list.");
      bmdlink* x = t; 
      t = pred_item(t, 0);
      if (t)
	{
	  if (t->ptr1 == x)
	    t->ptr1 = 0;
	  else
	    t->ptr2 = 0;
	}
      else
	h = 0;
      GenPtr p = x->e;
      delete x;
      count--;
      return p;
    }

  void reverse()
    {
      bmdlink* x = h;
      h = t;
      t = x;
    }
  void conc(bmdlist& l)
    {
      if (h == nil)
	{
	  h = l.h; 
	  t = l.t; 
	}
      else
	{
	  if (t->ptr1 == 0)
	    t->ptr1 = l.h;
	  else
	    t->ptr2 = l.h;
	  if (l.h)
	    {
	      if (l.h->ptr2 == 0)
		l.h->ptr2 = t;
	      else
		l.h->ptr1 = t;
	      t = l.t;
	    } 
	}
      count += l.count;
      l.h = l.t = 0;
      l.count = 0;
    }

  virtual ~bmdlist()  { }
  void clear()
    {
      if (h == 0)
	return;
      bmdlink* it = h;
      bmdlink* p = h;
      for(int i = 0; i < count; i++)
	{
	  bmdlink* x = it;
	  it = next_item(it, p);
	  if (p != x)
	    delete p;
          p = x;
	}
      delete p;
      count = 0;
      h = t = 0;
    }
  void swap(bmdlist& l)
    {
      bmdlink* tmp;
      int tmp1;
      tmp = l.h;
      l.h = h;
      h = tmp;
      tmp = l.t;
      l.t = t;
      t = tmp;
      tmp1 = l.count;
      l.count = count;
      count = tmp1;
    }
};

template<class E> 
class bmlist : public bmdlist 
{
public:
  bmlist_item push(const E& x) { return bmdlist::push(leda_copy(x)); }
  bmlist_item append(const E& x) { return bmdlist::append(leda_copy(x)); }
  const E& pop() { return LEDA_CONST_ACCESS(E, bmdlist::pop()); } 
  const E& pop_back() { return LEDA_CONST_ACCESS(E, bmdlist::pop_back()); } 
  const E& del(bmlist_item it) { return LEDA_CONST_ACCESS(E, bmdlist::del_item(it)); }
  const E& contents(bmlist_item it) { return LEDA_CONST_ACCESS(E, bmdlist::entry(it)); }
};

//------------------------------------------------------------------------------
//   Modified from iteration.h.
//   myforall_items(it,...)
//   myforall_rev_items(it,...)
//   myforall(x,...)
//   myforall_rev(x,...)
//------------------------------------------------------------------------------
// "quebra-gralho"
// can't use, for example, myforall_items twice in the same block!!

#define myforall_items(it,S)\
bmlist_item __p__ = (S).first_item();\
for(bmlist_item __it__ = (S).first_item();\
__it__, it = __it__;\
__it__ = (S).next_item(__it__, __p__), __p__ = it)

#define myforall_rev_items(it,S)\
bmlist_item __s__ = (S).last_item();\
for(bmlist_item __it__ = (S).last_item();\
__it__, it = __it__;\
__it__ = (S).pred_item(__it__, __s__), __s__ = it)

template<class var_type>
inline bool myLedaLoopInf(bmlist<var_type>& S, var_type& x, bmlist_item __it__, bmlist_item& __a__) 
{ __a__ = __it__; if (__it__) x = (S).contents(__it__); return __it__; }
#define myforall(x,S)\
bmlist_item __pp__ = (S).first_item(), __a__;\
for(bmlist_item __it__ = (S).first_item();\
myLedaLoopInf(S, x, __it__, __a__);\
__it__ = (S).next_item(__it__, __pp__), __pp__ = __a__)

#define myforall_rev(x,S)\
bmlist_item __ss__ = (S).last_item(), __b__;\
for(bmlist_item __it__ = (S).last_item();\
myLedaLoopInf(S, x, __it__, __b__);\
__it__ = (S).pred_item(__it__, __ss__), __ss__ = __b__)

#endif
