//-----------------------------------------------------------------------------------
// An implementation of Shih & Hsu's algorithm to test planarity.
// Modules: sh_planar.h, sh_impl.h, sh_test.c, sh_aux.c, sh_embed.c, sh_obstr.c,
//          dfspos.c, dfspos.h, bmplanar.h.
//
// Description:
//
// This module contains all auxiliary functions and procedures used in
// sh_test.c, sh_obstr.c and sh_embed.c.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#include "sh_impl.h"

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//-----------------------------------------------------------------------------------
// Returns the active c-node of RBC list containing "u".
// Traverses only nodes with label_b <= dfspos_num[w].
// Let z be the first element of RBC.
// All traversed nodes, but z, are inserted in "nl".
// Pre-conditions:
// - "u" is not the 1st element in RBC list;
// - ptritem[u] is the pointer to the item containing "u";
// - "u" is a node that is an ancestor of a neighbor of w;
// - state[v] == NOT_VISITED for each node v in RBC list that was not
//   traversed.
//-----------------------------------------------------------------------------------
node sh_planar_impl::find_active_cnode(node u, node w, list<node>& nl)
{
  list<node> traversed_nodes_in_RBC;


#ifdef _DEBUG_MODE_
//  cout<<"\nfind_list("<<dfspos_num[u]<<"): ";
  if (!is_cnode(parent[u]))
    error_handler(1, "-> Program terminated! (find_active_cnode01)");
#endif

  if (state[u] != NOT_VISITED)
    {
#ifdef _DEBUG_MODE_
      if (is_cnode(parent[parent[u]]))
	{
/*
	  cout<<"u="<<dfspos_num[u]<<",pu="<<dfspos_num[parent[u]]
              <<",ppu="<<dfspos_num[parent[parent[u]]];
*/
	  error_handler(1, "-> Program terminated! (find_active_cnode02)");
	}
#endif
      return parent[u];
    }

  bmlist_item it1 = ptritem[u];
#ifdef _DEBUG_MODE_
  if (it1 == nil)
    error_handler(1, "-> Program terminated! (find_active_cnode00)");
#endif

  state[u] = VISITED_IN_RBC;
  traversed_nodes_in_RBC.append(u);
  bmlist_item it = search_RBC(1, it1, w, traversed_nodes_in_RBC);
  if (it == nil) it = search_RBC(0, it1, w, traversed_nodes_in_RBC);
#ifdef _DEBUG_MODE_
  if (it == nil)
    error_handler(1, "-> Program terminated! (find_active_cnode03)");
#endif

  node v = *(const node*)(&it->e);
  node cnode;
  if (it->ptr1 != nil && it->ptr2 != nil)
    cnode = parent[v]; // path compressed;
  else
    cnode = activecnode[it];
#ifdef _DEBUG_MODE_
  if (cnode == nil)
    error_handler(1, "-> Program terminated! (find_active_node04)");
#endif

  node first = RBC[cnode].contents(RBC[cnode].first_item());
  // path compression;
  forall(v, traversed_nodes_in_RBC)
    {
      if (v != first)
	{
#ifdef _DEBUG_MODE_
//	  cout<<dfspos_num[v]<<",";
#endif
	  if (v != u)
	    nl.append(v);
	  parent[v] = cnode;
	}
      else state[v] = NOT_VISITED;
    }
#ifdef _DEBUG_MODE_
//  cout<<"cnode="<<dfspos_num[cnode]<<"\n";
#endif
  return cnode;
}
bmlist_item sh_planar_impl::search_RBC(int dir, bmlist_item it, node w,
                                       list<node>& traversed_nodes_in_RBC)
{
  if (it != nil && (it->ptr1 == nil || it->ptr2 == nil)) // 1st or last item in RBC;
    return it;
  bmlist_item prev = it, aux1 = it->ptr1, aux2 = it->ptr2;
  if (dir == 1) leda_swap(aux1, aux2);
  it = aux1;
  node u = *(const node*)(&it->e);
  int b = label_b[u];
  while (it != nil && (b <= dfspos_num[w] || dir != 1) && state[u] == NOT_VISITED)
    {
      aux1 = it->ptr1;
      if (aux1 == prev) aux1 = it->ptr2;
      prev = it;
      it = aux1;
      u = *(const node*)(&prev->e);
      state[u] = VISITED_IN_RBC;
      traversed_nodes_in_RBC.append(u);
      if (it != nil)
	{
	  u = *(const node*)(&it->e);
	  b = label_b[u];
	}
    }
  if (it == nil)
    return prev;
  else if (state[u] != NOT_VISITED || (it->ptr1 == nil || it->ptr2 == nil))
    return it;
  else
    return nil;
}

 //-----------------------------------------------------------------------------------
// Returns true for all edges e in T_0,
//         false otherwise.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::is_T0_edge(edge e)
{
  edge e1 = T0_edge_in[target(e)];
  if (e1 != nil && source(e1) == source(e) && target(e1) == target(e))
    return true;
  e1 = T0_edge_in[source(e)];
  return (e1 != nil && target(e1) == source(e) && source(e1) == target(e));
}

//-----------------------------------------------------------------------------------
// Returns true for all edges e in G - T_0,
//         false otherwise.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::is_back_edge(edge e)
{
  if (e == nil)
    return false;
  return (!is_T0_edge(e));
}

//-----------------------------------------------------------------------------------
// Every c-node u is marked with a negative value of dfspos_num[u].
// Returns true if u is a c-node,
//         false otherwise.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::is_cnode(node u)
{
  if (u == nil)
    return false;
  return (dfspos_num[u] < 0);
}

//-----------------------------------------------------------------------------------
// Returns the active c-node of u, if u belong to a c-node,
//         nil otherwise.
//-----------------------------------------------------------------------------------
node sh_planar_impl::active_cnode_of(bool calculating_obstruction, node u)
{
  node cnode = u;
  if (!is_cnode(u))
    cnode = parent[u];
  if (!is_cnode(cnode))
    return nil;
/*
#ifdef _DEBUG_MODE_
  cout<<"\nactive_cnode_of("<<dfspos_num[u]<<") :";
  node k=cnode;
  while(is_cnode(k))
    {
      cout<<dfspos_num[k]<<",";
      k = parent[k];
    }
  cout<<"\n";
#endif
*/
  if (calculating_obstruction)
    while (is_cnode(parent[cnode]))
      cnode = parent[cnode];
  else
    if (is_cnode(parent[cnode]))
      error_handler(1, "-> Program terminated! (active_cnode_of00)");
  return cnode;
}

//-----------------------------------------------------------------------------------
// Returns the least common ancestor lca between t1 and t2.
// If lca is a c-node, then it returns an active c-node.
// Precondition:
// - t1 and t2 are terminal nodes in the same 2-connected component.
//-----------------------------------------------------------------------------------
node sh_planar_impl::lca_between_term_nodes(
					    node t1,
					    node t2)
{
  node lca = last_visited[t1];
  if (dfspos_num[last_visited[t2]] < dfspos_num[lca])
    lca = last_visited[t2];
#ifdef _DEBUG_MODE_
  node lca2 = lca_between(t1, t2, parent);
  cout << "\nlca=" << dfspos_num[lca] << ", lca2=" << dfspos_num[lca2] << "\n";
  if (lca != lca2)
    error_handler(1, "-> Program terminated! (lcatnodes00)");
#endif
  return lca;
}

//-----------------------------------------------------------------------------------
// Returns the least common ancestor lca between t1 and t2 in T represented by p.
// If lca is a c-node, then it returns an active c-node.
// Precondition:
// - T is connected and t1 and t2 are in T.
//-----------------------------------------------------------------------------------
node sh_planar_impl::lca_between(
				 node t1,
				 node t2,
				 node_map<node> p)
{
  if (is_cnode(t1))
    t1 = p[active_cnode_of(false, t1)];
  if (is_cnode(t2))
    t2 = p[active_cnode_of(false, t2)];
  if (dfspos_num[t1] > dfspos_num[t2])
    leda_swap(t1, t2);
  list<node> nl;
  while (dfspos_num[t1] < dfspos_num[t2])
    {
      nl.push(t1);
      t1 = p[t1];
    }
  node u = nil;
  if (nl.size() > 0)
    u = nl.pop();
  // **** see gprof/gw/gteste20480.gw;
  while (t2 != u && t2 != t1 && dfspos_num[t2] < dfspos_num[t1])
    {
      nl.push(t2);
      t2 = p[t2];
    }
  if (t2 == u || t2 == t1)
    return t2;
  return nl.pop();
}

//-----------------------------------------------------------------------------------
// Returns the last p-node in path from t1 to t2 in T.
// Preconditions:
// - t2 is an ancestor of t1 in T;
// - there is a p-node in path from t1 to t2 in T that is not a c-node.
//-----------------------------------------------------------------------------------
node sh_planar_impl::last_pnode(node t1, node t2)
{
  if (t1 == t2)
    {
      if (!is_cnode(t1))
	return t1;
      else
	return nil;
    }
  list<node> S;
  node t = t1;
  while (t != nil && t != t2)
    {
      S.push(t);
      t = parent[t];
    }
  if (t == nil)
    return nil;
  t = t2;
/*
#ifdef _DEBUG_MODE_
  int count = 0;
#endif
*/
  while (is_cnode(t))
    {
/*
#ifdef _DEBUG_MODE_
      if (++count > 2)
	error_handler(1, "-> Program terminated! (cnodes01)");
#endif
*/
      if (S.empty())
	return nil;
      t = S.pop();
    }
  return t;
}

//-----------------------------------------------------------------------------------
// Returns in sorted_nodes the nodes u of G sorted in increasing order by value[u].
// Precondition:
// - value[u] <= n for all nodes u of G, where n is the number of nodes in G.
//-----------------------------------------------------------------------------------
void sh_planar_impl::sort_nodes_increasing_order(
						 graph_type& G,
						 node_map<int>& value,
						 array<node>& sorted_nodes)
{
  // Counting sort;

  int n = G.number_of_nodes();
  array<int> C(1, n);
  for (int i = 1; i <= n; i++)
    C[i] = 0;

  array<node> A(1, n);
  int j = 0;
  node u;
  forall_nodes(u, G)
    A[++j] = u;

  for (int i = 1; i <= n; i++)
    C[value[A[i]]]++;

  for (int i = 2; i <= n; i++)
    C[i] += C[i-1];

  for (int i = n; i > 0; i--)
    {
      sorted_nodes[C[value[A[i]]]] = A[i];
      C[value[A[i]]]--;
    }
}

//-----------------------------------------------------------------------------------
// Updates label_b[t].
// All traversed nodes in children_in_T0[t] that are not a descendant of t in
// current T are removed from children_in_T0[t].
// Preconditions:
// - children_in_T0[t] is a list with t's children in T_0 in decreasing order of
//   label_b;
// - t is not a c-node.
//-----------------------------------------------------------------------------------
void sh_planar_impl::update_label_b(node t)
{
  if (t == nil)
    return;
  label_b[t] = largest_neighbor[t];
  if (embed)
    node_label_b[t] = t; // needed to calculate obstruction edges if G is not planar;
  node v;
  if (!children_in_T0[t].empty())
    v = children_in_T0[t].contents(children_in_T0[t].first());
  else
    return;
  while (v != nil)
    {
      // or v is not a descendant of t in T,
      // or v is a child of t in T,
      // or v is a child of a c-node whose parent is t;
      if (parent[v] != nil && is_cnode(parent[v]) && parent[parent[v]] == t)
	{
	  v = parent[v]; // active c-node;
	  break;
	}
      if (parent[v] == t)
	break;

      // v became sibling of t;
      children_in_T0[t].del(children_in_T0[t].first());
      if (!children_in_T0[t].empty())
	v = children_in_T0[t].contents(children_in_T0[t].first());
      else v = nil;
    }
  if (v != nil && label_b[t] < label_b[v])
    {
      label_b[t] = label_b[v];
      if (embed)
	node_label_b[t] = node_label_b[v]; // needed to calculate obstruction edges;
    }
}

//-----------------------------------------------------------------------------------
// Returns a p-node v, if it exists, s.t. v is not in T_t, v is a
// descendant of a node in path Q_linha from parent[t] to a child of w, upward in T,
// and label_b[v] > dfspos_num[w].
// Returns nil otherwise.
//-----------------------------------------------------------------------------------
node sh_planar_impl::find_node_with_label_b_greater_than_dfs_w(
							       bool save_last_node_traversed,
							       graph_type& G,
							       node w,
							       node t)
{
  node
    result = nil,
    v_pred = t,
    v = parent[t];

  list<node> nl, nl2; // to restore parent, label_b and node_label_b;
  map<node, node> p, nb;
  map<node, int> b;

  p[t] = parent[t]; // saves info;
  nl.append(t);
  parent[t] = nil;
  while (v != w)
    {
      if (is_cnode(v))
	{
	  v = active_cnode_of(false, v);
	  node u;
	  myforall(u, RBC[v])
	    if (label_b[u] > dfspos_num[w] && u != t && u != parent[v])
	      {
		result = u;
		break;
	      }
	}
      else
	{
	  nl2.append(v);
	  b[v] = label_b[v]; // saves infos;
          if (embed) // needed to calculate obstruction edges;
	    nb[v] = node_label_b[v];
	  update_label_b(v);
	  if (label_b[v] > dfspos_num[w])
	    {
	      result = v;
	      break;
	    }
	}
      v_pred = v;
      v = parent[v];
      p[v_pred] = parent[v_pred];
      nl.append(v_pred);
      parent[v_pred] = nil;
    }

  // restores parent;
  forall(v, nl)
    parent[v] = p[v];

  if (result != nil)
    return result;

  // restores info if result is nil;
  forall(v, nl2)
    {
      label_b[v] = b[v];
      if (embed) // needed to calculate obstruction edges;
	node_label_b[v] = nb[v];
    }

  // to update label_b of w correctly in future iteration
  // (v_pred is the child of w in T that belongs to Q_linha);
  if (save_last_node_traversed && v_pred != t)
    last_node_in_Q_linha = v_pred; // see sets_info_for_new_cnode;
  else
    last_node_in_Q_linha = nil;
  return nil;
}

//-----------------------------------------------------------------------------------
// Tests each c-node counter in path from terminal "t1" to a p-node "t2", until
// it detects an obstruction or a possible obstruction.
// Returns true if there is an obstruction, false otherwise.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::test_obstruction_from_terminal_node(
							 graph_type& G,
							 node w,
							 node t1,
							 node u)
{
  node v = t1, pred_v = nil;
  while (v != u)
    {
      if (is_cnode(v))
	{
	  v = active_cnode_of(false, v);
	  node jl, jr;
	  if (test_cnode_counter(G, v, w, pred_v, nil, jl, jr))
	    {
	      if (embed)
		obstruction_edges_cnode_counter(G, v, w, jl, jr, t1, nil);
	      return true;
	    }
	  if (cnode_of_possible_k33_obstruction != nil)
	    break;
	  v = parent[v];
	  if (v == u)
	    break;
	}
      pred_v = v; // never a c-node;
      v = parent[v];
    }
  return false;
}

//-----------------------------------------------------------------------------------
// Returns true if testing counter[cnode] indicates an obstruction, false otherwise.
// Executes set_possible_K33_obstruction if any possible K33 obstruction detected.
// Preconditions:
// - the first element in list RBC[cnode] is parent[cnode];
// - there exists at least one node u in RBC[cnode] s.t. label_b[u] > dfspos_num[w]
//   and u != parent[cnode];
// - if w1 != nil then w1 is a node in RBC[cnode] s.t. label_b[w1] > dfspos_num[w]
//   and there exists a path from w1 to w that doesn't contain any node in
//   the 2-connected component represented by cnode different from w1;
// - idem for w2;
// - w1 != w2 and w1 and w2 are children of cnode.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::test_cnode_counter(
					graph_type& G,
					node cnode,
					node w,
					node w1,
					node w2,
					node& jl, node& jr)
{
#ifdef _DEBUG_MODE_
  if (parent[cnode] != RBC[cnode].contents(RBC[cnode].first_item()))
    error_handler(1, "-> Program terminated! (iscnodeobstr00)");
#endif
  // no chance for cnode to be an obstruction;
  if (RBC[cnode].size() <= 2)
    return false;

#ifdef _DEBUG_MODE_
  cout << "                    Testing c-node=" << dfspos_num[cnode] << "; counter=" 
       << counter[cnode] << ": w1 = ";
  if (w1 != nil)
    cout << dfspos_num[w1];
  else
    cout << "nil";
  cout << ", w2 = ";
  if (w2 != nil)
    cout << dfspos_num[w2];
  else
    cout << "nil";
  cout << "\n";
  cout << "                      {" << dfspos_num[parent[cnode]] << ": {";
  node v;
  myforall(v, RBC[cnode])
    {
      // if v == parent[cnode] or v == w1 or v == w2 (and v != nil)
      // then label_b[v] >= dfspos_num[w];
      if ((v == parent[cnode] && label_b[v] > dfspos_num[w])
          || v == w1 || v == w2)
	cout << "(" << dfspos_num[v] << "," 
             << dfspos_num[w] << "/" << label_b[v] << "),";
      else
	cout << "(" << dfspos_num[v] << "," << label_b[v] << "),";
    }
  cout << "} ";
#endif

  bmlist_item
    it1 = RBC[cnode].first_item(),
    it1l = RBC[cnode].cyclic_pred(it1, nil),
    it1r = RBC[cnode].cyclic_succ(it1, nil);
 
  jl = RBC[cnode].contents(it1l);
  jr = RBC[cnode].contents(it1r);
  node ul = nil, ur = nil;
  int count = 0;

  // goes to the left;
  bmlist_item
    a,
    s = it1;
  while (label_b[jl] <= dfspos_num[w])
    {
#ifdef _DEBUG_MODE_
      cout << dfspos_num[jl] << ", ";
#endif
      if (label_b[jl] == dfspos_num[w])
	{
	  ul = jl;
	  count++;
	}
#ifdef _DEBUG_MODE_
      else
	{
	  cout<<"w="<<dfspos_num[w]<<",jl="<<dfspos_num[jl]<<",b="<<label_b[jl]<<"\n";
	  error_handler(1, "-> Program terminated! (iscnodeobstr-X)");
	}
#endif
      a = it1l;
      it1l = RBC[cnode].cyclic_pred(it1l, s);
      s = a;
      jl = RBC[cnode].contents(it1l);
    }

  // goes to right;
  bmlist_item p = it1;
  while (label_b[jr] <= dfspos_num[w])
    {
#ifdef _DEBUG_MODE_
      cout << dfspos_num[jr] << ", ";
#endif
      if (label_b[jr] == dfspos_num[w])
	{
	  ur = jr;
	  count++;
	}
#ifdef _DEBUG_MODE_
      else
	{
	  cout<<"w="<<dfspos_num[w]<<",jr="<<dfspos_num[jr]<<",b="<<label_b[jr]<<"\n";
	  error_handler(1, "-> Program terminated! (iscnodeobstr-X)");
	}
#endif
      a = it1r;
      it1r = RBC[cnode].cyclic_succ(it1r, p);
      p = a;
      jr = RBC[cnode].contents(it1r);
    }

  if (jl == w1 || jl == w2)
    count++;
  if (jr != jl && (jr == w1 || jr == w2))
    count++;
#ifdef _DEBUG_MODE_
  cout << "(counter=" << count << ")}\n";
  if (count > counter[cnode])
    error_handler(1, "-> Program terminated! (iscnodeobstr02)");
  if (w1 == parent[cnode] || w2 == parent[cnode])
    error_handler(1, "-> Program terminated! (iscnodeobstr01)");
#endif
  if (count != counter[cnode])
    return true;

  node u = nil;
  if (ul != nil && ur != nil)
    u = jl; // or u = jr;
  else if (ul != nil && w1 != nil && w1 != jl)
    {
      u = jl;
      ur = w1;
    }
  else if (ur != nil && w1 != nil && w1 != jr)
    {
      u = jr;
      ul = w1;
    }
  if (cnode_of_possible_k33_obstruction == nil && u != nil)
    set_possible_K33_obstruction(cnode, u, ul, ur);
  return false;
}

//-----------------------------------------------------------------------------------
// Sets info about a possible obstruction (type 2) detected in c-node cnode.
// (cnode_parent_of_possible_k33_obstruction is set)
//-----------------------------------------------------------------------------------
void sh_planar_impl::set_possible_K33_obstruction(
						  node cnode,
						  node u,
						  node ul,
						  node ur)
{
  // K_33: X = {w, parent[cnode], u}
  //       Y = {ul, ur, ?};
  obstruction_nodes.clear();
  obstruction_nodes.append(u);
  obstruction_nodes.append(ul);
  obstruction_nodes.append(ur);
  cnode_of_possible_k33_obstruction = cnode;
#ifdef _DEBUG_MODE_
  cout<<"possible_k33="<<dfspos_num[cnode]<<"\n";
#endif
}

//-----------------------------------------------------------------------------------
// Calculates RBC for new_cnode. 
// Precondition:
// - terminal_nodes.size() == 1 or 2.
//-----------------------------------------------------------------------------------
void sh_planar_impl::calculate_new_RBC(
				       graph_type& G,
				       node new_cnode,
				       node w,
				       list<node>& terminal_nodes)
{
  /*
#ifdef _DEBUG_MODE_
  cout << "          P = ";
#endif
  */

  switch(terminal_nodes.size())
    {
      case 1: 
	{
	  node t1 = terminal_nodes.pop();
          calc_new_RBC_from_terminal_node(new_cnode, w, t1, w, RBC[new_cnode]);
	}
	break;

      case 2: 
	{
	  node
	    m = nil,
	    t1 = terminal_nodes.pop(),
	    t2 = terminal_nodes.pop(),
	    aux1 = t1, aux2 = t2;
	  // makes t1 with lower dfspos_num than t2;
	  if (is_cnode(aux1))
	    {
	      aux1 = active_cnode_of(false, aux1);
	      aux1 = parent[aux1];
	    }
	  if (is_cnode(aux2))
	    {
	      aux2 = active_cnode_of(false, aux2);
	      aux2 = parent[aux2];
	    }
	  if (dfspos_num[aux1] > dfspos_num[aux2])
	    leda_swap(t1, t2);

	  m = lca_between_term_nodes(t1, t2);
	  node t12 = last_pnode(t1, m), t22 = last_pnode(t2, m);
	  bmlist<node> auxnodelist;
	  calc_new_RBC_from_terminal_node(new_cnode, w, t1, t12, auxnodelist);
	  calc_new_RBC_from_terminal_node(new_cnode, w, t2, t22, RBC[new_cnode]);

	  // adds m to RBC[new_cnode];
          if (is_cnode(m))
	    {
	      m = active_cnode_of(false, m);
	      parent[t12] = parent[t22] = new_cnode;
	      update_label_b(t12);
	      if (label_b[t12] > dfspos_num[w])
		ptritem[t12] = auxnodelist.append(t12);
	      update_label_b(t22);
	      if (label_b[t22] > dfspos_num[w])
		ptritem[t22] = RBC[new_cnode].append(t22);
	      add_old_cnode_RBC_to_new_RBC(m, new_cnode, w, t22, t12, RBC[new_cnode]);
	      parent[m] = new_cnode;
	    }
	  else
	    {
	      parent[m] = new_cnode;
	      update_label_b(m);
	      if (label_b[m] > dfspos_num[w])
		ptritem[m] = RBC[new_cnode].append(m);
	    }
	  if (label_b[m] > label_b[new_cnode])
	    {
	      label_b[new_cnode] = label_b[m];
              if (embed) // needed to calculate obstruction edges;
		node_label_b[new_cnode] = node_label_b[m];
	    }
	  auxnodelist.reverse();
	  RBC[new_cnode].conc(auxnodelist);
	}
	break;

#ifdef _DEBUG_MODE_
      default: 
	error_handler(1, "-> Program terminated! (calculate_new_RBC00)");
#endif
    }
}

//-----------------------------------------------------------------------------------
// For all nodes v in path from t1 to t2 in T,
// sets parent[v] to be new_cnode, updates label_b[v] and adds v to nodelist if
// (updated) label_b[v] > dfspos_num[w].
// Preconditions:
// - t1 is a descendant of t2 in T;
// - t2 is not a c-node.
//-----------------------------------------------------------------------------------
void sh_planar_impl::calc_new_RBC_from_terminal_node(
						     node new_cnode,
						     node w,
						     node t1,
						     node t2,
						     bmlist<node>& nodelist)
{
  node 
    t = t1,
    pred_t = nil,
    parent_t = nil;
  while (t != t2)
    {
      /*
#ifdef _DEBUG_MODE_
        // path P;
	cout << dfspos_num[t] << ", ";
#endif
      */
      parent_t = parent[t];
      if (is_cnode(t)) 
	{
	  t = active_cnode_of(false, t);
	  add_old_cnode_RBC_to_new_RBC(t, new_cnode, w, pred_t, nil, nodelist);
	  parent_t = parent[t];
	  parent[t] = new_cnode;
	  if (label_b[t] > label_b[new_cnode])
	    {
	      label_b[new_cnode] = label_b[t];
              if (embed) // needed to calculate obstruction edges;
		node_label_b[new_cnode] = node_label_b[t];
	    }
	}
      else
	{
	  parent[t] = new_cnode;
	  update_label_b(t);
	  if (label_b[t] > dfspos_num[w])
	    ptritem[t] = nodelist.append(t);
	  if (label_b[t] > label_b[new_cnode])
	    {
	      label_b[new_cnode] = label_b[t];
	      if (embed) // needed to calculate obstruction edges;
		node_label_b[new_cnode] = node_label_b[t];
	    }
	}
      if (!is_cnode(t))
        pred_t = t; // pred_t is never a c-node;
      t = parent_t;
    }
}

//-----------------------------------------------------------------------------------
// Concatenates RBC[oldcnode] to nodelist.
// Preconditions:
// - the first element in list RBC[oldcnode] is parent[oldcnode];
// - if w1 != nil then w1 is a node in RBC[oldcnode] s.t. label_b[w1] > dfspos_num[w]
//   and there exists a path from w1 to w that doesn't contain any node in the
//   2-connected component represented by oldnode different from w1. 
// - idem for w2.
//-----------------------------------------------------------------------------------
void sh_planar_impl::add_old_cnode_RBC_to_new_RBC(
						  node oldcnode,
						  node newcnode,
						  node w,
						  node w1,
						  node w2,
						  bmlist<node>& nodelist)
{
  // compress RBC[oldcnode]:
  // removes all nodes v in RBC[oldcnode] s.t. label_b[v] < dfspos_num[w]
  // or v = parent[oldcnode].
  bmlist_item 
    it1 = RBC[oldcnode].first_item(),
    it1l = RBC[oldcnode].cyclic_pred(it1, nil),
    it1r = RBC[oldcnode].cyclic_succ(it1, nil);

  node 
    jl = RBC[oldcnode].contents(it1l),
    jr = RBC[oldcnode].contents(it1r),
    ul = nil, ur = nil;

#ifdef _DEBUG_MODE_
  if (parent[oldcnode] != RBC[oldcnode].contents(it1))
    error_handler(1, "-> Program terminated! (add_old_cnode_RBC_to_new_RBC00)");
#endif

  // goes to the left;
  while (label_b[jl] == dfspos_num[w] && jl != w1 && jl != w2)
    {
      if (ul == nil)
	ul = jl;
      bmlist_item aux1 = it1l;
      it1l = RBC[oldcnode].cyclic_pred(it1l, it1);
      jl = RBC[oldcnode].contents(it1l);
      RBC[oldcnode].del(aux1); 
    }

  // goes to right;
  while (label_b[jr] == dfspos_num[w] && jr != w1 && jr != w2)
    {
      if (ur == nil)
	ur = jr;
      bmlist_item aux1 = it1r;
      it1r = RBC[oldcnode].cyclic_succ(it1r, it1);
      jr = RBC[oldcnode].contents(it1r);
      RBC[oldcnode].del(aux1);
    }

  RBC[oldcnode].del(RBC[oldcnode].first_item()); 

  // endpoint to correctly concatentates with RBC[new_cnode];
  node first = w1;
  if (w1 == nil)
    if (ul != nil)
      first = jl;
    else
      first = jr;

  // note that oldcnode may have flipped;
  if (RBC[oldcnode].contents(RBC[oldcnode].last_item()) == first)
     RBC[oldcnode].reverse();
#ifdef _DEBUG_MODE_
     // is w1 the 2nd element in the list RBC[oldcnode]? 
  else if (RBC[oldcnode].contents(RBC[oldcnode].first_item()) != first)
    error_handler(1, "-> Program terminated! (add_old_cnode_RBC_to_new_RBC01)");
#endif
  // removes w1 and w2 from RBC[oldcnode];
  if (w1 != nil)
    RBC[oldcnode].del(RBC[oldcnode].first_item());
  if (w2 != nil)
    {
#ifdef _DEBUG_MODE_
      if (RBC[oldcnode].contents(RBC[oldcnode].last_item()) != w2)
	error_handler(1, "-> Program terminated! (add_old_cnode_RBC_to_new_RBC02)");
#endif
      RBC[oldcnode].del(RBC[oldcnode].last_item());
    }
  nodelist.conc(RBC[oldcnode]);
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
