//-----------------------------------------------------------------------------------
// Precondition:
//   - uncomment 
//       #define _DEBUG_MODE_
//     in file sh_impl.h.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------
#include <LEDA/graph.h>
#include <LEDA/graph_gen.h>
#include <LEDA/graph_misc.h>
#include <LEDA/plane_graph_alg.h>
#include <LEDA/string.h>
#include <stdio.h>
#include "sh_planar.h"
#include <LEDA/IO_interface.h>
#include <string.h>

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//#define _PRINT_LATEX_

/*
 * Reads a parameterized graph <int, string> from a file.
 * Returns 0 if successful, 
 *         > 0 otherwise. 
 */
int read_graph(graph_type& G, string gw_file)
{
  int error = G.read(gw_file);
  switch (error)
    {
    case 0:
      cout << "** Graph in file " << gw_file 
           << " successfully loaded.\n";
      break;
    case 1: 
      cout << "* Error: File " << gw_file << " not found!\n";
      break;
    case 2: 
      cout << "* Error: The edge and node parameter"
           <<" types in file " 
           << gw_file << " do not match!\n";
      break;
    case 3: 
      cout << "* Error: File " << gw_file 
           << " with invalid format!\n";
      break;
    }
  return error;
}

bool run_test(graph_type& G, int embed, list<edge>& K)
{
  G.make_map();

  float T1 = used_time();
  bool planar;
  if (embed == 1)
    {
      cout << "\n** Running SH_PLANAR - TEST+EMBEDDING\n";
      planar = SH_PLANAR(G, K, true);
    }
  else
    {
      cout << "\n** Running SH_PLANAR - TEST\n";
      planar = SH_PLANAR(G);
    }
  T1 = used_time(T1);
  printf("...Elapsed time: %.4f\n", T1);

  if (embed == 1)
    if (!planar)
      {
	cout << "   G is NOT planar!\n";
	if (!CHECK_KURATOWSKI(G, K))
	  error_handler(1, "-> Program terminated! (Wrong kuratowski!)");
      }
    else
      {
	cout << "   G is planar.\n";
	if (Genus(G) != 0)
	  error_handler(1, "-> Program terminated! (Wrong embedding!)");
      }
  return planar;
}

int main(int argc, char **argv)
{
  graph_type G;

  // run program on specified graph:
  if (argc >= 2)
    {
#ifdef _PRINT_LATEX_
      cout << "\\section{"
           << argv[1]
           << "}\n\\begin{figure}[htbp]\n\\begin{center}\n"
           << "\\includegraphics{../hsushih/ps/"
           << argv[1]
           << ".eps}\n\\end{center}\n\\end{figure}\n"
           << "\\footnotesize\n\\begin{verbatim}\n";
#endif

      //int embed = atoi(argv[1]);
      char **endptr;
      int embed = strtol(argv[1], endptr, 10);
      //cout << "endptr=" << **endptr << "\n";
      if (**endptr != '\0')
	goto _USAGE_MESSAGE_;

      if (argc == 3)
	{
	  if (read_graph(G, argv[2]))
	    return 1;

	  if (!G.is_bidirected())
	    {
	      cout << "Warning: Making G bidirected.\n";
	      G.make_bidirected();
	    }
	  list<edge> K;
	  bool planar = run_test(G, embed, K);
	  if (!planar)
	    {
	      edge_map<bool> in_K(G);
	      edge e;
	      forall_edges(e, G)
		in_K[e] = false;
	      forall(e, K)
		in_K[e] = true;
	      list<edge> el;
	      forall_edges(e, G)
		if (!in_K[e])
		  el.append(e);
	      G.del_edges(el);
	      cout << "G is NOT planar.\n";
	    }
	  else cout << "G is planar.\n";
	  if (embed == 1)
	    {
	  strcat(argv[2], ".out");
	  G.write(argv[2]);
	  cout << "Ouput:" << argv[2] << "\n";
	    }
#ifdef _PRINT_LATEX_
	  cout << "\\end{verbatim}\n\\normalsize\n";
#endif
	}
      else // run program on random graphs:
	{
	  cout << "\n** Testing with random: \n\n";
	  while (true)
	    {
	      int n = rand_int(10, 100), 
		//n = 1500, //m=3 * n - 6;
		m = rand_int(n / 2, 3 * n - 6),
		t = rand_int(1, 2);
	      if (t == 1)
		random_graph(G, n, m, true, true, true);
	      else
	        random_planar_graph(G, n, m);
	      G.make_bidirected();
	      G.write("graph.gw");
	      cout << "   (n, m) = (" << n << ", " << m << ")\n";

	      list<edge> K;
	      run_test(G, embed, K);
	      G.clear();
	    }
	}
    }
  else
    {
    _USAGE_MESSAGE_:
      cout << "Usage: sh_planar <int e> [graph.gw]\n"
	   << "     e: if e = 1 then Test+Embedding,"
	   << " else Test.\n\n";
    }
  return 0;
}
//***********************************************************************************
