//-----------------------------------------------------------------------------------
// An implementation of Shih & Hsu's algorithm to test planarity.
// Modules: sh_planar.h, sh_impl.h, sh_test.c, sh_aux.c, sh_embed.c, sh_obstr.c, 
//          dfspos.c, dfspos.h, bmplanar.h.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#include "sh_impl.h"

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//-----------------------------------------------------------------------------------
// Embeds all back-edges from T's root and all remaining edges in G that weren't
// embedded yet.
// Preconditions:
// - G is a graph with n nodes;
// - G is biconnected.
//-----------------------------------------------------------------------------------
void sh_planar_impl::embed_root(graph_type& G, int n)
{
  if (n <= 2)
    return;
  list<node> traversed_nodes;
  edge e;
  node 
    r = node_with_dfspos[n],
    u;
#ifdef _DEBUG_MODE_
  cout << "   " << dfspos_num[r] << ": ";
#endif
  state[r] = VISITED;
  forall_out_edges(e, r)
    {
      u = target(e);
      if (dfspos_num[u] < dfspos_num[r] && is_back_edge(e))
	{
	  back_edge_out[u] = G.reverse(e);
	  list_back_edges[r].append(G.reverse(e));
	  while(state[u] != VISITED)
	    {
#ifdef _DEBUG_MODE_
	      cout << dfspos_num[u] << "(" << dfspos_num[parent[u]] << "), ";
#endif
	      // path compression;
	      if (is_cnode(parent[u]))
		find_active_cnode(u, r, traversed_nodes);

	      state[u] = VISITED;
	      traversed_nodes.append(u);
	      u = parent[u];
	    }
	}
    }
  forall(u, traversed_nodes)
    state[u] = NOT_VISITED;

  array<edge> back_edge;
  int nH = sort_back_edges_by_dfs(r, r, list_back_edges[r], back_edge);
 
 bmlist<edge> el1, l1;
  for(int i = 1; i <= nH; i++)
    {
      e = back_edge[i];
      if (e != nil)
	{
	  node pred_u = source(e), u = parent[pred_u];
	  embed_list[r].push(e);
	  embed_list[r].append(G.reverse(e));
	  while (state[pred_u] == NOT_VISITED)
	    {
	      state[pred_u] = VISITED;
	      if (is_cnode(u))
		{
		  u = active_cnode_of(false, u);
		  l1.conc(embed_list[u]);
		  u = parent[u];
		}
	      else
		{
		  l1.append(G.reverse(T0_edge_in[pred_u]));
		  l1.append(T0_edge_in[pred_u]);
		}
	      pred_u = u;
	      u = parent[u];
	    }
	  l1.conc(el1);
	  l1.swap(el1);
	}
    }
  el1.conc(embed_list[r]);
  el1.swap(embed_list[r]);

  list<edge> final_emb_list;
  myforall(e, embed_list[r])
    final_emb_list.append(e);

#ifdef _DEBUG_MODE_
  { // due to myforall macro;
    map<node, list<edge> > el;
    list<node> nl;
    node_array<bool> marked(G, false);
    myforall(e, embed_list[r])
      {
	node w = source(e);
	if (!marked[w])
	  {
	    marked[w] = true;
	    nl.append(w);
	  }
	el[w].append(e);
      }
    cout << "\n       * FINAL Embed List: ";
    cout << "[" << embed_list[r].size() << " / " 
	 << G.number_of_edges() << "]\n";
    node k;
    forall(k, nl)
      {
	cout << "         " << dfspos_num[k] << ": ";
	forall(e, el[k])
	  cout << dfspos_num[target(e)] << ", ";
	cout << "\n";
      }
    cout << "         List of edges: ";

    { // due to myforall macro;
      myforall(e, embed_list[r])
	cout << "(" << dfspos_num[source(e)] << "," <<
	dfspos_num[target(e)] << ") ";
      cout << "\n";
    }
  }
#endif

  G.sort_edges(final_emb_list);
}

//-----------------------------------------------------------------------------------
// Calculates a partial embedding for the 2-connected component represented by
// new_cnode (embed_list[new_cnode]).
// Preconditions:
// - u is not a root of T;
// - state[u] == NOT_VISITED for all nodes u in T_w;
// - for all nodes u in T_w, has_back_edge[u] == false;
// - T_w is biconnected.
//-----------------------------------------------------------------------------------
void sh_planar_impl::calculate_partial_embedding(
						 graph_type& G,
						 node w,
						 node new_cnode,
						 list<edge>& list_back_edges,
						 list<node>& terminal_nodes)
{
  list<node> traversed_nodes, list_representants;
  switch (terminal_nodes.size())
    {
#ifdef _DEBUG_MODE_
    case 0:
      error_handler(1, "-> Program terminated! (calcpartialemb00!)");
      break;
#endif
    case 1:
      {
        map<node, node> backedge_representant;
	node term = terminal_nodes.contents(terminal_nodes.first());

	// marks as VISITED all nodes in the boundary cycle;
	mark_path_in_T(term, w, backedge_representant, traversed_nodes);
	map<node, list<edge> > b_edges_repres;
        b_edges_repres = group_back_edges_by_repr(G, list_back_edges,
						  backedge_representant,
						  traversed_nodes,
						  list_representants);
        list<node> to_embed_later;
	to_embed_later = embed_upward_T(true, term, w, G, w, b_edges_repres, traversed_nodes,
			   embed_list[new_cnode]);

	// embeds all nodes in to_embed_later;
	node t;
	forall(t, to_embed_later)
	  {
	    bmlist<edge> el1;
	    embed_back_edges(false, G, t, traversed_nodes, b_edges_repres[t],
			     el1);
	    el1.reverse();
	    el1.conc(embed_list[new_cnode]);
	    el1.swap(embed_list[new_cnode]);

	    embed_list[new_cnode].conc(list_back_edges_out_w);
	  }
      }
      break;
    case 2: 
      {
	map<node, node> backedge_representant;
	node
	  term1 = terminal_nodes.contents(terminal_nodes.first()),
	  term2 = terminal_nodes.contents(terminal_nodes.last()),
	  t1 = term1, t2 = term2,
	  m = lca_between_term_nodes(t1, t2);
	// makes term1 with lowest dfspos_num
	// (as in calculate_new_RBC);
	if (is_cnode(t1))
	  t1 = parent[t1];
	if (is_cnode(t2))
	  t2 = parent[t2];
#ifdef _DEBUG_MODE_
	if (is_cnode(t1) || is_cnode(t2))
	  error_handler(1, "-> Program terminated! (calcpartialemb01!)");
#endif
	if (dfspos_num[t1] > dfspos_num[t2])
	  leda_swap(term1, term2);

	t1 = last_pnode(term1, m);

	mark_path_in_T(term2, w, backedge_representant, traversed_nodes);
	mark_path_in_T(term1, t1, backedge_representant, traversed_nodes);

	map<node, list<edge> > b_edges_repres;
	b_edges_repres = group_back_edges_by_repr(G, list_back_edges,
						  backedge_representant,
						  traversed_nodes,
						  list_representants);
	list<node> to_embed_later;
	to_embed_later = embed_upward_T(true, term2, w, G, w, b_edges_repres, traversed_nodes,
			   embed_list[new_cnode]);
	node t;
	forall(t, to_embed_later)
	  embed_back_edges(true, G, t, traversed_nodes, b_edges_repres[t],
			   embed_list[new_cnode]);

	if (t1 != m)
	  {
	    bmlist<edge> el1;
	    embed_back_edges(false, G, t1, traversed_nodes, b_edges_repres[t1],
			     el1);
	    el1.reverse();
	    el1.conc(embed_list[new_cnode]);
	    el1.swap(embed_list[new_cnode]);

	    embed_list[new_cnode].conc(list_back_edges_out_w);
	  }
	bmlist<edge> el1;
	embed_upward_T(false, term1, t1,
		       G, w, b_edges_repres, traversed_nodes, el1);
	el1.reverse();
	el1.conc(embed_list[new_cnode]);
	el1.swap(embed_list[new_cnode]);

	embed_list[new_cnode].conc(list_back_edges_out_w);
      }
      break;
    default: // graph is not planar (abort procedure);
      return;
    }
  // restores auxiliary variables;
  node u;
  forall(u, traversed_nodes)
    state[u] = NOT_VISITED;
  forall(u, list_representants)
    has_back_edge[u] = false;

#ifdef _DEBUG_MODE_
  map<node, list<edge> > el;
  list<node> nl;
  node_array<bool> marked(G, false);
  edge e;
  myforall(e, embed_list[new_cnode])
    {
      node u = source(e);
      if (!marked[u])
	{
	  marked[u] = true;
	  nl.append(u);
	}
      el[u].append(e);
    }
  cout << "       * Embed List (cnode=" << dfspos_num[new_cnode] << "): ";
  cout << "[" << embed_list[new_cnode].size() << " / " 
       << G.number_of_edges() << "]\n";
  node k;
  forall(k, nl)
    {
      cout << "         " << dfspos_num[k] << ": ";
      forall(e, el[k])
	cout << dfspos_num[target(e)] << ", ";
      cout << "\n";
    }
  cout << "         List of edges: ";

  { // due to myforall macro;
    myforall(e, embed_list[new_cnode])
      cout << "(" << dfspos_num[source(e)] << "," <<
      dfspos_num[target(e)] << ") ";
    cout << "\n";
  }
#endif
}

//-----------------------------------------------------------------------------------
// For all node u in path from t to w upward in T,
// initializes backedge_representant[u] = u,
// marks u as VISITED and appends u to list traversed_nodes.
//-----------------------------------------------------------------------------------
void sh_planar_impl::mark_path_in_T(node t, node w,
				    map<node, node>& backedge_representant,
				    list<node>& traversed_nodes)
{
  state[w] = VISITED;
  backedge_representant[w] = w;
  traversed_nodes.append(w);
#ifdef _DEBUG_MODE_
  cout << "                    Boundary cycle: ";
  cout << dfspos_num[w] << ", ";
#endif
  node u = t;
  while (state[u] == NOT_VISITED)
    {
#ifdef _DEBUG_MODE_
      cout << dfspos_num[u] << ", ";
#endif
      state[u] = VISITED;
      traversed_nodes.append(u);
      backedge_representant[u] = u;
      u = parent[u];
    }
#ifdef _DEBUG_MODE_
  cout << "\n";
#endif
}

//-----------------------------------------------------------------------------------
// For each back-edge e in T_w (list_back_edges) calculates
// backedge_representant[source(e)] and returns all back-edges grouped by
// backedge_representant.
// Marks as VISITED all nodes that are a representant of a back-edge.
// Appends all nodes marked as VISITED in list traversed_nodes.
// Precondition:
// - for all nodes u in T_w, has_back_edge[u] == false.
//-----------------------------------------------------------------------------------
map<node, list<edge> > sh_planar_impl::group_back_edges_by_repr(
					      graph_type& G,
					      list<edge>& list_back_edges,
					      map<node, node>& backedge_representant,
					      list<node>& traversed_nodes,
					      list<node>& list_representants)
{
  list<node> nl;
  edge e;
  forall(e, list_back_edges)
    {
      list<node> S;
      node
	u = source(e),
	pnode = u; // pnode is never a c-node;
      while (state[u] == NOT_VISITED)
	{
	  if (!is_cnode(u))
	    pnode = u;
	  if (state[u] == NOT_VISITED)
	    {
	      state[u] = VISITED;
	      nl.append(u);
	      S.push(u);
	    }
	  u = parent[u];
	}
      node repr;
      if (!is_cnode(backedge_representant[u]))
	repr = backedge_representant[u];
      else
	{
	  repr = pnode;
	  traversed_nodes.append(repr);
	}
      forall(u, S)
	{
/*
#ifdef _DEBUG_MODE_
	  if (is_cnode(u) && is_cnode(parent[u]) && is_cnode(parent[parent[u]]))
	    error_handler(1, "-> Program terminated! (sortlistbackedges00)");
#endif
*/
	  if (is_cnode(u) && is_cnode(parent[u]))
	    backedge_representant[u] = backedge_representant[parent[u]];
	  else
	    backedge_representant[u] = repr;
	}
      if (!has_back_edge[repr])
	{
	  has_back_edge[repr] = true;
	  list_representants.append(repr);
	}
    }
#ifdef _DEBUG_MODE_
  cout << "                    Back-edges and their representants: ";
#endif
  map<node, list<edge> > list_edges;
  forall(e, list_back_edges)
    {
#ifdef _DEBUG_MODE_
      cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)]
	   << ")[" << dfspos_num[backedge_representant[source(e)]] << "], ";
#endif
      node v = source(e);
      list_edges[backedge_representant[v]].append(e);
    }
#ifdef _DEBUG_MODE_
  cout << "\n";
#endif

  node u;
  forall(u, nl)
    state[u] = NOT_VISITED;
  forall(u, list_representants)
    state[u] = VISITED;

    return list_edges;
}

//-----------------------------------------------------------------------------------
// Embeds all edges in path P' from t1 to t2 and all back-edges with
// representant in P' (emb_list).
// emb_back_edges_out_w is false only in case of 2 terminal
// nodes, for one of the two terminals (see calculate_partial_embedding).
// Returns an ordered list of all representants to be embedded later
// (see add_old_cnode_to_embedding).
// Precondition:
// - if u is a node in P' and e is a back-edge with representant u then e is in list
//   b_edges_repres[u].
//-----------------------------------------------------------------------------------
list<node> sh_planar_impl::embed_upward_T(
					  bool emb_back_edges_out_w,
					  node t1,
					  node t2,
					  graph_type& G,
					  node w,
					  map<node, list<edge> >& b_edges_repres,
					  list<node>& traversed_nodes,
					  bmlist<edge>& emb_list)
{
  list<node> to_embed_later;
  node u = t1, pred_u = nil;
  while (pred_u != t2)
    {
      if (is_cnode(u))
	{
	  node
	    f = pred_u,
	    oldcnode = active_cnode_of(false, u);
	  add_old_cnode_to_embedding(emb_back_edges_out_w, G, w, oldcnode, f,
				     b_edges_repres, traversed_nodes, to_embed_later,
				     emb_list);
	  u = parent[oldcnode];
	  if (u == t2)
	    return to_embed_later;
	}
      else
	{
	  if (pred_u != nil)
	    {
	      emb_list.push(G.reverse(T0_edge_in[pred_u]));
	      if (u != w)
		emb_list.push(T0_edge_in[pred_u]);
	      else
		emb_list.append(T0_edge_in[pred_u]);
	    }
	}
      if (has_back_edge[u] && u != t2)
	embed_back_edges(emb_back_edges_out_w, G, u, traversed_nodes,
			 b_edges_repres[u], emb_list);
      pred_u = u;
      u = parent[u];
    }
  return to_embed_later;
}

//-----------------------------------------------------------------------------------
// Moves embedding of oldcnode to emb_list - note that oldcnode may
// flip - and embeds all edges in path from w (starting with a back-edge from w) and
// ending in a node in RBC[oldcnode].
// All representants to be embedded later is ordered and inserted in to_embed_later.
// emb_back_edges_out_w is false only in case of 2 terminal
// nodes, for one of the two terminals (see calculate_partial_embedding).
// Preconditions:
// - for a back-edge e, if u is representant of source(e) and  u in RBC[oldcnode],
//   then e is in list b_edges_repres[u];
// - the first element in list RBC[oldcnode] is parent[oldcnode];
// - if u != nil then u is a node in RBC[oldcnode] s.t. label_b[u] > dfspos_num[w]
//   and there exists a path from u to w that doesn't contain any node in the
//   2-connected component represented by oldcnode, except u.
//-----------------------------------------------------------------------------------
void sh_planar_impl::add_old_cnode_to_embedding(
						bool emb_back_edges_out_w,
						graph_type& G,
						node w,
						node oldcnode,
						node u,
						map<node, list<edge> >& b_edges_repres,
						list<node>& traversed_nodes,
						list<node>& to_embed_later,
						bmlist<edge>& emb_list)
{
  bmlist_item 
    it = RBC[oldcnode].first_item(),
    itl = RBC[oldcnode].cyclic_pred(it, nil),
    itr = RBC[oldcnode].cyclic_succ(it, nil);
  node 
    jl = RBC[oldcnode].contents(itl),
    jr = RBC[oldcnode].contents(itr);

  list<node> list_nodesl, list_nodesr;
  // goes to the left;
  bmlist_item
    aux,
    s = it;
  while (label_b[jl] <= dfspos_num[w])
    {
#ifdef _DEBUG_MODE_
      if (jl == u)
	error_handler(1, "-> Program terminated! (addoldcnodeemb00a)");
#endif
      if (label_b[jl] == dfspos_num[w])
	list_nodesl.append(jl);
      aux = itl;
      itl = RBC[oldcnode].cyclic_pred(itl, s);
      s = aux;
      jl = RBC[oldcnode].contents(itl);
    }
  // goes to the right;
  bmlist_item p = it;
  while (label_b[jr] <= dfspos_num[w])
    {
#ifdef _DEBUG_MODE_
      if (jr == u)
	error_handler(1, "-> Program terminated! (addoldcnodeemb00b)");
#endif
      if (label_b[jr] == dfspos_num[w])
	list_nodesr.append(jr);
      aux = itr;
      itr = RBC[oldcnode].cyclic_succ(itr, p);
      p = aux;
      jr = RBC[oldcnode].contents(itr);
    }
  // checks if need to flip oldcnode;
  // u == nil when oldcnode is a terminal node;
  bool flipped = ((list_nodesl.size() > 0 && (jl == u || u == nil))
                   || (jr != u && u != nil));
  if (flipped)
    list_nodesl.swap(list_nodesr);
  // embeds all back-edges in oldcnode;
  list_nodesr.reverse_items();
  node t;
  forall(t, list_nodesr)
    embed_back_edges(emb_back_edges_out_w, G, t, traversed_nodes,
		     b_edges_repres[t], emb_list);
  if (flipped)
    embed_list[oldcnode].reverse();

  embed_list[oldcnode].conc(emb_list);
  embed_list[oldcnode].swap(emb_list);

  to_embed_later.conc(list_nodesl, LEDA::before);
}

//-----------------------------------------------------------------------------------
// Embeds all edges in a path from w, starting in a back-edge in list_back_edges
// and ending in repr.
// Inserts all embedded edges in list emb_list.
// Appends all nodes marked as VISITED in list traversed_nodes.
// emb_back_edges_out_w is false only in case of 2 terminal
// nodes, for one of the two terminals (see calculate_partial_embedding).
//-----------------------------------------------------------------------------------
void sh_planar_impl::embed_back_edges(
				      bool emb_back_edges_out_w,
				      graph_type& G,
				      node repr,
				      list<node>& traversed_nodes,
				      list<edge>& list_back_edges,
				      bmlist<edge>& emb_list)
{
  if (list_back_edges.empty())
    return;
  edge e;
  bmlist<edge> el1, l1, wl1;
  node w = target(list_back_edges.contents(list_back_edges.first()));

  array<edge> back_edge;
  int n = sort_back_edges_by_dfs(w, repr, list_back_edges, back_edge);

  for(int i = 1; i <= n; i++)
    {
      e = back_edge[i];
      if (e != nil)
	{
	  if (emb_back_edges_out_w)
	    wl1.append(G.reverse(e));
	  else
	    list_back_edges_out_w.push(G.reverse(e));
	}
    }
  for(int i = n; i >= 1; i--)
    {
      e = back_edge[i];
      if (e != nil)
	{
	  l1.push(e);
	  node
	    pred_u = source(e),
	    u = parent[pred_u];
	  while (state[pred_u] == NOT_VISITED)
	    {
	      state[pred_u] = VISITED;
	      traversed_nodes.append(pred_u);
	      if (is_cnode(u))
		{
		  u = active_cnode_of(false, u);
		  embed_list[u].conc(l1);
		  embed_list[u].swap(l1);
		  u = parent[u];
		}
	      else
		{
		  l1.push(T0_edge_in[pred_u]);
		  l1.push(G.reverse(T0_edge_in[pred_u]));
		}
	      pred_u = u;
	      u = parent[u];
	    }
	  el1.conc(l1);
	}
    }
  el1.conc(emb_list);
  el1.swap(emb_list);
  emb_list.conc(wl1);
}

//-----------------------------------------------------------------------------------
// Sorts all back-edges with representant repr by depth first search traversal in
// (T_repr - P), denoted as T_v^*.
//-----------------------------------------------------------------------------------
int sh_planar_impl::sort_back_edges_by_dfs(
					   node w,
					   node repr,
					   list<edge>& list_back_edges,
					   array<edge>& back_edge)
{
  // constructs a DFS tree of T_v^* to sort back-edges to embed;
  graph D;
  list<node> list_nodes, list_cnodes;
  map<node, node> node_in_D, node_in_G;
  edge e;
  node_in_D[repr] = D.new_node();
  node_in_G[node_in_D[repr]] = repr;
  forall(e, list_back_edges)
    {
      node u = source(e), pred_u = nil;
      while(state[u] == NOT_VISITED)
	{
	  if (is_cnode(u))
	    {
	      u = active_cnode_of(false, u);
	      if (state[u] == NOT_VISITED)
		list_cnodes.append(u);
	    }
	  if (state[u] == NOT_VISITED)
	    {
	      state[u] = VISITED;
	      list_nodes.append(u);
	      node_in_D[u] = D.new_node();
	      node_in_G[node_in_D[u]] = u;
	    }
	  if (pred_u != nil)
	    D.new_edge(node_in_D[u], node_in_D[pred_u]);
	  pred_u = u;
	  u = parent[u];
	}
      if (pred_u != nil)
	D.new_edge(node_in_D[u], node_in_D[pred_u]);
    }
  node v;
  forall(v, list_nodes)
    state[v] = NOT_VISITED;

  // creates  edges from c-nodes;
  map<node, bool> is_in_D;
  forall(v, list_cnodes)
    {
      node u;
      myforall(u, RBC[v])
	is_in_D[u] = false;
      list<edge> el;
      forall_out_edges(e, node_in_D[v])
	{
          is_in_D[node_in_G[target(e)]] = true;
	  el.append(e);
	}
      D.del_edges(el);
      myforall_rev(u, RBC[v])
	if(is_in_D[u])
	  D.new_edge(node_in_D[v], node_in_D[u]);
    }

  // orders back-edges by DFS number;
  node_array<int> dfspos(D, 0);
  DFS_POS(D, dfspos);
  int tot = D.number_of_nodes();
  back_edge.resize(1, tot);
  for (int i = 1; i <= tot; i++)
    back_edge[i] = nil;
  forall(e, list_back_edges)
    {
      v = source(e);
      back_edge[dfspos[node_in_D[v]]] = e;
    }
  return tot;
}

//-----------------------------------------------------------------------------------
// Algebric criteria to check the plane map...
//-----------------------------------------------------------------------------------
void sh_planar_impl::check_embedding(graph_type& G)
{
  if (Genus(G) != 0)
    error_handler(1, "-> Program terminated! (embedding00!)");

  // checks if f = m - n + 2;

  // piece of code from _g_misc.c (from LEDA source directory);
  edge_array<bool> considered(G,false);
  int fc = 0;
  edge e;
  forall_edges(e,G)
    {
      if (!considered[e])
	{
	  // trace the face to the left of e
	  edge e1 = e;
	  do {
	    considered[e1] = true;
	    e1 = G.face_cycle_succ(e1);
	  }
	  while (e1 != e);
	  fc++;
	}
    }

  int
    m = G.number_of_edges() / 2,
    n = G.number_of_nodes();
  if (fc != m - n + 2)
    error_handler(1, "-> Program terminated! (f != m - n + 2)");
}
//-----------------------------------------------------------------------------------
