//-----------------------------------------------------------------------------------
// Header for modules: sh_test.c, sh_embed.c, sh_obstr.c.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#ifndef _SH_IMPL_H
#define _SH_IMPL_H

#include <LEDA/string.h>
#include <LEDA/graph.h>
#include <LEDA/node_array.h>
#include <LEDA/node_map.h>
#include <LEDA/list.h>
#include <LEDA/graphwin.h>
#include <LEDA/bigfloat.h>
#include <LEDA/graph_gen.h>
#include <LEDA/graph_misc.h>
#include <LEDA/plane_graph_alg.h>
#include "dfspos.h"
#include "bmlist.h"

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//-----------------------------------------------------------------------------------
// DEBUGGING...
//-----------------------------------------------------------------------------------
//#define _DEBUG_MODE_
#ifdef _DEBUG_MODE_
  #define _TESTING_MAPS_ONLY_
  typedef GRAPH<int, string> graph_type;
#endif

#ifndef _DEBUG_MODE_
  typedef graph graph_type;
#endif

//-----------------------------------------------------------------------------------
// IMPLEMENTATION
//-----------------------------------------------------------------------------------
// "state";
enum { NOT_VISITED, VISITED, TERMINAL, VISITED_IN_RBC };

class sh_planar_impl
{
  //-------- member variables --------
private:
  // to avoid path compression of c-nodes;
  map<bmlist_item,node> activecnode;
  node_map<bmlist_item> ptritem;

  bool embed, biconnected;

  // DFS post order numbering of each node of T, where T is a DFS tree of G;
  node_map<int> dfspos_num;

  // if dfspos_num[u] == i then node_with_dfspos[i] == u;
  array<node> node_with_dfspos;

  // to help calculate an embedding or an obstruction;
  node_array<edge> T0_edge_in;
  node_array<edge> back_edge_out;
  
  node_map<node>
    // let T be a DFS tree of G,
    // for each node u in T, if u is not a root, u has a parent defined in T;
    parent,
    p0;            // saves initial DFS tree T_0 of G;

  // for each node u in T, 
  // largest_neighbor[u] = max{dfspos_num[v] : v is a neighbor of u in G};
  node_map<int> largest_neighbor;

  // for each node u in T,
  // label_b[u] = max{largest_neighbor[v] : v is a descendat of u in T_u}
  // where T_u is the subtree of T rooted at u;
  node_map<int> label_b; 

  // for each node u in T, node_label_b[u] = v
  // where v is a descendant of u in T and largest_neighbor[v] == label_b[u]
  // (it helps to find an obstruction in G, if G is not planar); 
  node_map<node> node_label_b;

  // to help find the lca between two terminal nodes;
  node_map<node> last_visited;

  // given w, for each terminal node u of w, neighbor_w_terminal[u] is 
  // a descendant of u that is a neighbor of w in G;
  node_map<node> neighbor_w_terminal;

  // for each node u in T, children is the list of u's children
  // ordered in decreasing order by label_b
  // (it helps to update label_b's in constant time);
  node_array<list<node> > children_in_T0;

  // to help search for terminal nodes and calculate an embedding of G if G is
  // planar (states: VISITED, NOT_VISITED, TERMINAL);
  node_map<int> state; 

  // the Representative Boundary Cycle for each c-node;
  map<node, bmlist<node> > RBC; 

  // for each (active) c-node d, counter[d] is the number of children of d
  // with a descendant that are neighbor of w in G;
  map<node, int> counter;    

  // list of nodes in an obstruction found in G if G is not planar
  // (it helps to calculate "obstruction_edges");
  list<node> obstruction_nodes;

  // list of edges in an obstruction found int G if G is not planar;
  list<edge> obstruction_edges;

  // auxiliary variable to help detecting obstruction;
  node cnode_of_possible_k33_obstruction;

  int total_cnodes;

  // for each node u in G, the algorithm calculates the
  // clockwise ordering of edges with source u around u, such that
  // G.sort_edges(embed_list) is a plane map, if it exists
  map<node, bmlist<edge> > embed_list;

  // for each 2-connected component represented by r,
  // list_back_edges[r] is the list of all back-edges in component r
  // (it helps to calculate an embedding of G, if G is planar);
  node_array<list<edge> > list_back_edges;

  // (it helps to calculate an embedding of G, if G is planar);
  node_array<bool> has_back_edge;

  // (it helps to calculate an embedding of G, if G is planar, in
  // case of 2 terminal nodes);
  bmlist<edge> list_back_edges_out_w;

  node last_node_in_Q_linha; // see sets_info_for_new_cnode;

#ifdef _DEBUG_MODE_
  int number_of_nodes_in_G;
#endif

  //-------- member functions --------
public:
  // constructor:
  sh_planar_impl(graph_type& G, bool emb, bool biconnect)
    {
      int n = G.number_of_nodes();
#ifdef _DEBUG_MODE_
      number_of_nodes_in_G = n; // used in sh_obstr.c;
#endif
      ptritem.init(G, nil);
      dfspos_num.init(G);
      node_with_dfspos.resize(1, n);
      parent.init(G, nil);
      largest_neighbor.init(G);
      label_b.init(G);
      node_label_b.init(G, nil);
      neighbor_w_terminal.init(G, nil);
      last_visited.init(G);
      children_in_T0.init(G);
      state.init(G, NOT_VISITED);
      total_cnodes = 0;
      T0_edge_in.init(G, nil);
      back_edge_out.init(G, nil);
      embed = emb;
      biconnected = biconnect;
      list_back_edges.init(G);
      has_back_edge.init(G, false);
      last_node_in_Q_linha = nil;
      if (embed)
	p0.init(G);
    }

private:
  //---- sh_aux.c:
  node find_active_cnode(node, node, list<node>&);
  bmlist_item search_RBC(int, bmlist_item, node, list<node>&);
  bool is_T0_edge(edge);
  bool is_back_edge(edge);
  bool is_cnode(node);
  node active_cnode_of(bool, node);
  node lca_between(node, node, node_map<node>);
  node lca_between_term_nodes(node, node);
  node last_pnode(node, node);
  void sort_nodes_increasing_order(graph_type&, node_map<int>&, array<node>&);
  void update_label_b(node);
  node find_node_with_label_b_greater_than_dfs_w(bool, graph_type&, node, node);
  bool test_obstruction_from_terminal_node(graph_type&, node, node, node);
  bool test_cnode_counter(graph_type&, node, node, node, node, node&, node&);
  void set_possible_K33_obstruction(node, node, node, node);
  void calculate_new_RBC(graph_type&, node, node, list<node>&);
  void calc_new_RBC_from_terminal_node(node, node, node, node, bmlist<node>&);
  void add_old_cnode_RBC_to_new_RBC(node, node, node, node, node, bmlist<node>&);

  //---- sh_test.c;
  void pre_processing(graph_type&);
  void find_terminal_nodes(graph_type&, node,list<node>&, node_array<list<node> >&);
  bool find_obstruction(graph_type&, node, list<node>&);
  void set_info_for_new_cnode(graph_type&, node, node, list<node> &);
public:
  bool is_planar(graph_type&);

private:
  //---- sh_embed.c;
  void embed_root(graph_type&, int);
  void calculate_partial_embedding(graph_type&, node, node, list<edge>&, list<node>&);
  void mark_path_in_T(node, node, map<node, node>&, list<node>&);
  map<node, list<edge> > group_back_edges_by_repr(
					         graph_type&, list<edge>&,
						 map<node, node>&, list<node>&,
						 list<node>&);
  list<node> embed_upward_T(
			    bool, node, node, graph_type&, node, map<node, list<edge> >&,
			    list<node>&, bmlist<edge>&);
  void add_old_cnode_to_embedding(
				  bool, graph_type&, node, node, node,
				  map<node, list<edge> >&, list<node>&, list<node>&,
				  bmlist<edge>&);
  void embed_back_edges(
			bool, graph_type&, node, list<node>&, list<edge>&,
			bmlist<edge>&);
  int sort_back_edges_by_dfs(node, node, list<edge>&, array<edge>&);
  void check_embedding(graph_type&);

  //---- sh_obstr.c:
  edge find_edge(graph_type&, node, node);
  bool list_edges_upward_T0(node, node);
  void extract_boundary_cycle(node, list<edge>&);
  void obstruction_edges_cnode_counter(
				       graph_type&, node, node, node, node, node, node);
  void obstruction_edges_possible_obstr_confirmed(graph_type&, node, node, node);
  void add_part_of_bc(node, node, node, node);
  void obstr_edges_terminal(graph_type&, node, node, node);
  void sort_by_label_b(node&, node&, node&);
  void obstr_edges_term_cnode(graph_type&, node, node);
  void obstr_edges_pnode(graph_type&, node, node);
  void obstruction_edges_K5(graph_type&, node, node, node, node, node);
  void obstruction_edges_countmin_1(graph_type&, node, node, node, node, node);
  void obstruction_edges_countmin_2_3(graph_type&, node, node, node, node, node, node, node);
  void obstruction_edges_T0(graph_type&, node, node, node, node, node);
  void calc_info_3_terminals(node&, node&, node&, int&, int&, node&, node&);
public:
  list<edge> obstruction();
};
#endif
