//-----------------------------------------------------------------------------------
// An implementation of Shih & Hsu's algorithm to test planarity.
// Modules: sh_planar.h, sh_impl.h, sh_test.c, sh_aux.c, sh_embed.c, sh_obstr.c, 
//          dfspos.c, dfspos.h, bmplanar.h.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#include "sh_impl.h"

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//-----------------------------------------------------------------------------------
// Returns the list of edges of an obstruction found in G, if G is not planar.
//-----------------------------------------------------------------------------------
list<edge> sh_planar_impl::obstruction()
{
  return obstruction_edges;
}

//-----------------------------------------------------------------------------------
// Returns edge "e" s.t. source(e) == u and target(e) == v, if it exists,
//         nil otherwise.
//-----------------------------------------------------------------------------------
edge sh_planar_impl::find_edge(graph_type& G, node u, node v)
{
  edge e;
  forall_out_edges(e, u)
    {
      node vv = target(e);
      if (vv == v)
	return e;
    }
  return nil;
}

//-----------------------------------------------------------------------------------
// Adds to "el" the edges in path from w1 to w2, upward in T0.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::list_edges_upward_T0(node w1, node w2)
{
  if (w1 == w2)
    return true;

  node u = w1;
  while (u != nil && u != w2)
    {
      edge e = T0_edge_in[u];
#ifdef _DEBUG_MODE_
      node v = p0[u];
      if (source(e) != v)
	{
	  cout<<"u="<<dfspos_num[u]<<",v="<<dfspos_num[v]<<"\n";
	  error_handler(1, "-> Program terminated! (list_edges_upward_T0-01)");
	}
#endif
      obstruction_edges.append(e);
      u = p0[u];
    }
  return (u == w2);
}

//-----------------------------------------------------------------------------------
// Adds boundary cycle of "cnode" to "list_edges".
// Precondition:
// - embed_list[cnode] is an embedding of the 2-connected component represented
//   by "cnode" with "RBC[cnode] contained in the boundary".
//-----------------------------------------------------------------------------------
void sh_planar_impl::extract_boundary_cycle(node cnode, list<edge>& list_edges)
{
  if (embed_list[cnode].size() <= 0)
    error_handler(1, "-> Program terminated! (bcycle00)");
  map<node, list<edge> > el;
  edge e;
  node u;
  myforall(e, embed_list[cnode])
    {
      u = source(e);
      el[u].append(e);
    }
  u = parent[cnode];
  e = el[u].contents(el[u].first());
  list_edges.append(e);
  u = target(e);
#ifdef _DEBUG_MODE_
  int count = 0;
#endif
  while (u != parent[cnode])
    {
#ifdef _DEBUG_MODE_
      if(++count > number_of_nodes_in_G)
	error_handler(1, "-> Program terminated! (bcycle01)");
#endif
      e = el[u].contents(el[u].first());
      list_edges.append(e);
      u = target(e);
    }
}

//-----------------------------------------------------------------------------------
// K33
// Pre-conditions:
// - for i = 1, 2, if t_i is not nil, then it is a terminal node that is descendant
// of cnode.
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_cnode_counter(
						     graph_type& G,
						     node cnode,
						     node w,
						     node jl,
						     node jr,
						     node t1,
						     node t2)
{
  // seachs for a node f in RBC[cnode] between jl and jr s.t.
  // f has a descendant that is a neighbor of w in G;
  bool flag = false;
  node v = nil, f = nil, t12 = nil, t22 = nil;
  if (t1 != nil && t1 != cnode)
    t12 = last_pnode(t1, cnode);
  if (t2 != nil && t2 != cnode)
    t22 = last_pnode(t2, cnode);
  myforall(v, RBC[cnode])
    {
      if (v == jl || v == jr)
	flag = true;
      else if (flag && (v == t12 || v == t22 || label_b[v] == dfspos_num[w]))
	{
	  f = v;
	  break;
	}
    }
  if (f == t12)
    node_label_b[f] = neighbor_w_terminal[t1];
  else if (f == t22)
    node_label_b[f] = neighbor_w_terminal[t2];
#ifdef _DEBUG_MODE_
  if (f == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-01)");
  cout<<"obstruction_edges_cnode_counter:\n";
  cout<<"  cnode="<<dfspos_num[cnode]<<",f="<<dfspos_num[f];
  cout<<"\n  jl="<<dfspos_num[jl]<<",b="<<label_b[jl]<<",nl="<<dfspos_num[node_label_b[jl]];
  cout<<"\n  jr="<<dfspos_num[jr]<<",b="<<label_b[jr]<<",nl="<<dfspos_num[node_label_b[jr]];
  cout<<"\n";
#endif
  node 
    w1 = node_with_dfspos[label_b[jl]],
    w2 = node_with_dfspos[label_b[jr]];
  if (dfspos_num[w2] < dfspos_num[w1])
    leda_swap(w1, w2);
  if (!list_edges_upward_T0(parent[cnode], w2))
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-02)");
  if (!list_edges_upward_T0(node_label_b[f], f))
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-03)");
  if (!list_edges_upward_T0(node_label_b[jl], jl))
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-04)");
  if (!list_edges_upward_T0(node_label_b[jr], jr))
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-05)");

  edge e = find_edge(G, node_label_b[f], w);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-06)");
  obstruction_edges.append(e);
  e = find_edge(G, node_label_b[jl], node_with_dfspos[label_b[jl]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-07)");
  obstruction_edges.append(e);
  e = find_edge(G, node_label_b[jr], node_with_dfspos[label_b[jr]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_cnode_counter-08)");
  obstruction_edges.append(e);

  extract_boundary_cycle(cnode, obstruction_edges);

#ifdef _DEBUG_MODE_
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_cnode_counter!)");
#endif
}

//-----------------------------------------------------------------------------------
// K33
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_possible_obstr_confirmed(
								graph_type& G,
								node w,
								node t,
								node v)
{
  node cnode = cnode_of_possible_k33_obstruction,
    f = obstruction_nodes.pop(),
    jl = obstruction_nodes.pop(),
    jr = obstruction_nodes.pop();
  if (label_b[jl] > dfspos_num[w])
    node_label_b[jl] = neighbor_w_terminal[t];
  if (label_b[jr] > dfspos_num[w])
    node_label_b[jr] = neighbor_w_terminal[t];
#ifdef _DEBUG_MODE_
  cout<<"obstruction_edges_possible_obstr_confirmed:\n";
  cout<<"  w="<<dfspos_num[w]<<",b="<<label_b[w]<<",nl="<<dfspos_num[node_label_b[w]];
  if (cnode == nil || v == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-01)");   
  cout<<"\n  cnode="<<dfspos_num[cnode]<<",f="<<dfspos_num[f]<<",v="<<dfspos_num[v];
  cout<<"\n  jl="<<dfspos_num[jl]<<",b="<<label_b[jl]<<",nl="<<dfspos_num[node_label_b[jl]];
  cout<<"\n  jr="<<dfspos_num[jr]<<",b="<<label_b[jr]<<",nl="<<dfspos_num[node_label_b[jr]];
  cout<<"\n  f="<<dfspos_num[f]<<",b="<<label_b[f]<<",nl="<<dfspos_num[node_label_b[f]];
  cout<<"\n  v="<<dfspos_num[v]<<",b="<<label_b[v]<<",nl="<<dfspos_num[node_label_b[v]];
  cout<<"\n";
#endif
  node 
    w1 = node_with_dfspos[label_b[v]],
    w2 = node_with_dfspos[label_b[f]];
  if (dfspos_num[w2] < dfspos_num[w1])
    leda_swap(w1, w2);
  if (!list_edges_upward_T0(w, w2))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-02)");
  if (!list_edges_upward_T0(node_label_b[f], f))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-03)");
  if (!list_edges_upward_T0(node_label_b[jl], jl))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-04)");
  if (!list_edges_upward_T0(node_label_b[jr], jr))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-05)");

  node m = lca_between(parent[cnode], v, p0);
  if (!list_edges_upward_T0(parent[cnode], m))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-06)");
  if (!list_edges_upward_T0(node_label_b[v], m))
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-07)");

  edge e = find_edge(G, node_label_b[v], node_with_dfspos[label_b[v]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-08)");
  obstruction_edges.append(e);
  e = find_edge(G, node_label_b[f], node_with_dfspos[label_b[f]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-09)");
  obstruction_edges.append(e);
  e = find_edge(G, node_label_b[jl], w);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-10)");
  obstruction_edges.append(e);
  e = find_edge(G, node_label_b[jr], w);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_possible_obstr_confirmed-11)");
  obstruction_edges.append(e);

  extract_boundary_cycle(cnode, obstruction_edges);

#ifdef _DEBUG_MODE_
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_possible_obstr_confirmed!)");
#endif
}

//-----------------------------------------------------------------------------------
// Adds part of the boundary cycle of "cnode" that contains w1 and doesn't contain
// path between w2 and w3 to "el".
//-----------------------------------------------------------------------------------
void sh_planar_impl::add_part_of_bc(node cnode, node w1, node w2, node w3)
{
  list<edge> bc, el1, el2;
  extract_boundary_cycle(cnode, bc);
  edge e;
  bool flag = false;
  int side = 0;
  forall(e, bc)
    {
      if (source(e) == w1)
	{
	  if (flag)
	    side = 1;
	  else
	    side = 2;
	}
      if (source(e) == w2 || source(e) == w3)
	flag = !flag;
      if (flag)
	el1.append(e);
      else
	el2.append(e);
    }
#ifdef _DEBUG_MODE_
  if (side == 0)
    error_handler(1, "-> Program terminated! (add_part_of_bc00)");
#endif
  if (side == 1)
    obstruction_edges.conc(el1, LEDA::after);
  else
    obstruction_edges.conc(el2, LEDA::after);
}


//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstr_edges_terminal(graph_type& G, node w, node t, node u)
{
  node mm = lca_between(node_label_b[t], neighbor_w_terminal[t], p0);
  if (!list_edges_upward_T0(node_label_b[t], mm))
    error_handler(1, "-> Program terminated! (obstr_edges_terminal01)");
  if (!list_edges_upward_T0(neighbor_w_terminal[t], mm))
    error_handler(1, "-> Program terminated! (obstr_edges_terminal02)");
  if (!list_edges_upward_T0(mm, u))
    error_handler(1, "-> Program terminated! (obstr_edges_terminal03)");

  edge e = find_edge(G, node_label_b[t], node_with_dfspos[label_b[t]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstr_edges_terminal04)");
  obstruction_edges.append(e);
  e = find_edge(G, neighbor_w_terminal[t], w);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstr_edges_terminal05)");
  obstruction_edges.append(e);
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void sh_planar_impl::sort_by_label_b(node &t1, node &t2, node &t3)
{
  if (label_b[t2] < label_b[t1])
    leda_swap(t1, t2);
  if (label_b[t3] < label_b[t1])
    leda_swap(t1, t3);
  if (label_b[t3] < label_b[t2])
    leda_swap(t2, t3);
}
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstr_edges_term_cnode(graph_type& G, node w, node t)
{
  node u = last_pnode(node_label_b[t], t),
    uu = last_pnode(neighbor_w_terminal[t], t);
  if (!list_edges_upward_T0(node_label_b[t], u))
    error_handler(1, "-> Program terminated! (obstr_edges_term_cnode01)");
  if (!list_edges_upward_T0(neighbor_w_terminal[t], uu))
    error_handler(1, "-> Program terminated! (obstr_edges_term_cnode02)");
  add_part_of_bc(t, parent[t], u, uu);

  edge e = find_edge(G, node_label_b[t], node_with_dfspos[label_b[t]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstr_edges_term_cnode03)");
  obstruction_edges.append(e);
  e = find_edge(G, neighbor_w_terminal[t], w);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstr_edges_term_cnode04)");
  obstruction_edges.append(e);
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstr_edges_pnode(graph_type& G, node p, node u)
{
  if (!list_edges_upward_T0(node_label_b[p], u))
    error_handler(1, "-> Program terminated! (obstr_edges_pnode01)");
  edge e = find_edge(G, node_label_b[p], node_with_dfspos[label_b[p]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstr_edges_pnode02)");
  obstruction_edges.append(e);
}

//-----------------------------------------------------------------------------------
// Pre-conditions:
// - t1, t2 and t3 are (p-nodes) terminals in the same c-node;
// - cnode is part of a K_5 obstruction.
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_K5(
					  graph_type& G,
					  node w,
					  node cnode,
					  node t1,
					  node t2,
					  node t3)
{
#ifdef _DEBUG_MODE_
  cout<<"obstruction_edges_K5:\n";
  cout<<"  w="<<dfspos_num[w]<<",cnode="<<dfspos_num[cnode];
  cout<<"\n  t1="<<dfspos_num[t1]<<",b="<<label_b[t1]<<",nl="<<dfspos_num[node_label_b[t1]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t1]];
  cout<<"\n  t2="<<dfspos_num[t2]<<",b="<<label_b[t2]<<",nl="<<dfspos_num[node_label_b[t2]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t2]];
  if (t3 != nil)
    cout<<"\n  t3="<<dfspos_num[t3]<<",b="<<label_b[t3]<<",nl="<<dfspos_num[node_label_b[t3]]
	<<",nt="<<dfspos_num[neighbor_w_terminal[t3]];
  cout<<"\n";
#endif
  if (t3 == nil)
    t3 = parent[cnode];
  int max = label_b[t1];
  if (max < label_b[t2])
    max = label_b[t2];
  if (max < label_b[t3])
    max = label_b[t3];
  if (!list_edges_upward_T0(w, node_with_dfspos[max]))
    error_handler(1, "-> Program terminated! (obstruction_edges_K5-01)");

  extract_boundary_cycle(cnode, obstruction_edges);

  obstr_edges_terminal(G, w, t1, t1);
  obstr_edges_terminal(G, w, t2, t2);
  if (t3 != parent[cnode])
    obstr_edges_terminal(G, w, t3, t3);
  else
    obstr_edges_pnode(G, parent[cnode], w);

#ifdef _DEBUG_MODE_
  edge e;
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_K5!)");
#endif
}

//-----------------------------------------------------------------------------------
// Pre-conditions:
// - t1, t2 and t3 are terminal nodes;
// - ti is p-node whose parent is cnode;
// - if t3 == nil, then parent[cnode] has a descendant v s.t.
//   label_b[v] > dfspos_num[w] and v is not a descendant of cnode;
// - cnode is part of a K_{3,3} obstruction.
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_countmin_1(
						  graph_type& G,
						  node w,
						  node cnode,
						  node t1,
						  node t2,
						  node t3)
{
#ifdef _DEBUG_MODE_
  cout<<"obstruction_edges_countmin_1:\n";
  cout<<"  w="<<dfspos_num[w]<<",cnode="<<dfspos_num[cnode];
  cout<<"\n  t1="<<dfspos_num[t1]<<",b="<<label_b[t1]<<",nl="<<dfspos_num[node_label_b[t1]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t1]];
  cout<<"\n  t2="<<dfspos_num[t2]<<",b="<<label_b[t2]<<",nl="<<dfspos_num[node_label_b[t2]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t2]];
  if (t3 != nil)
    cout<<"\n  t3="<<dfspos_num[t3]<<",b="<<label_b[t3]<<",nl="<<dfspos_num[node_label_b[t3]]
        <<",nt="<<dfspos_num[neighbor_w_terminal[t3]]<<"\n";
  else
    cout<<"\n  pcnode="<<dfspos_num[parent[cnode]]<<",b="<<label_b[parent[cnode]]
        <<",nl="<<dfspos_num[node_label_b[parent[cnode]]]<<"\n";
#endif
  if (t3 == nil)
    t3 = parent[cnode];
  sort_by_label_b(t1, t2, t3);
#ifdef _DEBUG_MODE_
  cout<<"min="<<label_b[t1]<<",max="<<label_b[t3]<<"\n";
#endif
  if (!list_edges_upward_T0(w, node_with_dfspos[label_b[t3]]))
    error_handler(1, "-> Program terminated! (obstruction_edges_countmin_1-01)");

  add_part_of_bc(cnode, t1, t2, t3);

  if (!list_edges_upward_T0(node_label_b[t1], t1))
    error_handler(1, "-> Program terminated! (obstruction_edges_countmin_1-02)");
  edge e = find_edge(G, node_label_b[t1], node_with_dfspos[label_b[t1]]);
  if (e == nil)
    error_handler(1, "-> Program terminated! (obstruction_edges_countmin_1-03)");
  obstruction_edges.append(e);

  if (t2 != parent[cnode])
    obstr_edges_terminal(G, w, t2, t2);
  else
    obstr_edges_pnode(G, parent[cnode], w);

  if (t3 != parent[cnode])
    obstr_edges_terminal(G, w, t3, t3);
  else
    obstr_edges_pnode(G, parent[cnode], w);

#ifdef _DEBUG_MODE_
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_countmin_1!)");
#endif
}

//-----------------------------------------------------------------------------------
// Pre-conditions:
// - t1, t2 and t3 are terminal nodes;
// - t1 is terminal p-node,
//   t1 = min{lca_in_T0(t1,t2,),lca_in_T0(t1,t3),lca_in_T0(t2,t3)}
//   and t1 is a "child" of cnode;
// - if t3 == nil, then v is a descendant in T of a node in path from
//   parent[parent[cnode]] to a child of w s.t. label_b[v] > dfspos_num[w];
//   label_b[v] > dfspos_num[w] and v is not a descendant of cnode;
// - cnode is part of a K_{3,3} obstruction.
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_countmin_2_3(
						    graph_type& G,
						    node w,
						    node cnode,
						    node t1,
						    node t2,
						    node t3,
						    node q,
						    node v)
{
#ifdef _DEBUG_MODE_
  cout<<"obstruction_edges_countmin_2_3:\n";
  cout<<"  w="<<dfspos_num[w]<<",cnode="<<dfspos_num[cnode]<<",q="<<dfspos_num[q];
  cout<<"\n  t1="<<dfspos_num[t1]<<",b="<<label_b[t1]<<",nl="<<dfspos_num[node_label_b[t1]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t1]];
  cout<<"\n  t2="<<dfspos_num[t2]<<",b="<<label_b[t2]<<",nl="<<dfspos_num[node_label_b[t2]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t2]];
  if (t3 != nil)
    cout<<"\n  t3="<<dfspos_num[t3]<<",b="<<label_b[t3]<<",nl="<<dfspos_num[node_label_b[t3]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t3]];
  else
    cout<<"\n  pcnode="<<dfspos_num[parent[cnode]];
  if (v != nil)
    cout<<"\n  v="<<dfspos_num[v]<<",b="<<label_b[v]
        <<",nl="<<dfspos_num[node_label_b[v]]<<"\n";
#endif
  node w1 = t1, w2 = t2, w3 = t3;
  if (t3 == nil)
    w3 = v;
  sort_by_label_b(w1, w2, w3);
  if (!list_edges_upward_T0(node_with_dfspos[label_b[w1]], node_with_dfspos[label_b[w3]]))
    error_handler(1, "-> Program terminated! (obstruction_edges_countmin_2_3-01)");

  w2 = last_pnode(t2, cnode);
  if (w2 != q)
    add_part_of_bc(cnode, q, t1, w2);
  else
    add_part_of_bc(cnode, q, t1, parent[cnode]);

  obstr_edges_terminal(G, w, t1, t1);

  obstr_edges_terminal(G, w, t2, w2);

  if (t3 != nil)
    {
      node u = lca_between(node_label_b[t3], q, p0);
      // "u" can be an ancestor of q when t3 is not descendant of cnode;
      obstr_edges_terminal(G, w, t3, u);
      if (!list_edges_upward_T0(q, u))
	error_handler(1, "-> Program terminated! (obstruction_edges_countmin_2_3-02)");
    }
  else
    {
      obstr_edges_pnode(G, v, w);
      node mm = lca_between(v, parent[cnode], p0);
      if (!list_edges_upward_T0(parent[cnode], mm))
	error_handler(1, "-> Program terminated! (obstruction_edges_countmin_2_3-03)");
    }

#ifdef _DEBUG_MODE_
  edge e;
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_countmin_2_3!)");
#endif
}

//-----------------------------------------------------------------------------------
// K33 -- T0;
// Pre-conditions:
// - t1, t2 and t3 are terminal nodes;
// - if parent_cnode != nil then t3 == nil.
//-----------------------------------------------------------------------------------
void sh_planar_impl::obstruction_edges_T0(
					  graph_type& G,
					  node w,
					  node t1,
					  node t2,
					  node t3,
					  node v)
{
#ifdef _DEBUG_MODE_
  cout<<"obstruction_edges_T0:\n";
  cout<<"  w="<<dfspos_num[w];
  cout<<"\n  t1="<<dfspos_num[t1]<<",b="<<label_b[t1]<<",nl="<<dfspos_num[node_label_b[t1]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t1]];
  cout<<"\n  t2="<<dfspos_num[t2]<<",b="<<label_b[t2]<<",nl="<<dfspos_num[node_label_b[t2]]
      <<",nt="<<dfspos_num[neighbor_w_terminal[t2]];
  if (t3 != nil)
    cout<<"\n  t3="<<dfspos_num[t3]<<",b="<<label_b[t3]<<",nl="<<dfspos_num[node_label_b[t3]]
        <<",nt="<<dfspos_num[neighbor_w_terminal[t3]];
  if (v != nil)
    cout<<"\n  v="<<dfspos_num[v]<<",b="<<label_b[v]
        <<",nl="<<dfspos_num[node_label_b[v]]<<"\n";
#endif
  if (t3 == nil)
    t3 = v;
  node w1 = t1, w2 = t2, w3 = t3;
  sort_by_label_b(w1, w2, w3);
  if (!list_edges_upward_T0(node_with_dfspos[label_b[w1]], node_with_dfspos[label_b[w3]]))
    error_handler(1, "-> Program terminated! (obstruction_edges_T0-01)");

  w1 = t1;
  w2 = t2;
  w3 = t3;
  if (is_cnode(w1))
    w1 = parent[w1];
  if (is_cnode(w2))
    w2 = parent[w2];
  if (is_cnode(w3))
    w3 = parent[w3];
  node m1 = lca_between(w1, w2, p0),
    m2 = lca_between(w1, w3, p0),
    m3 = lca_between(w2, w3, p0);
  int min = dfspos_num[m1], max = dfspos_num[m1];
  if (dfspos_num[m2] < min)
    {
      leda_swap(t2, t3);
      min = dfspos_num[m2];
    }
  if (dfspos_num[m3] < min)
    {
      node k1 = t2, k2 = t3, k3 = t1;
      t1 = k1;
      t2 = k2;
      t3 = k3;
      min = dfspos_num[m3];
    }
  if (dfspos_num[m2] > max)
    max = dfspos_num[m2];
  if (dfspos_num[m3] > max)
    max = dfspos_num[m3];

#ifdef _DEBUG_MODE_
  cout<<"min="<<min<<",max="<<max<<"\n";
#endif
  obstr_edges_terminal(G, w, t1, node_with_dfspos[max]);

  if (t3 != v)
    obstr_edges_terminal(G, w, t3, node_with_dfspos[max]);
  else
    {
      node u = w;
      if (lca_between(v, node_with_dfspos[min], p0) == node_with_dfspos[min])
	{
	  // v is in the same c-node of node_with_dfspos[min];
	  node cnode = active_cnode_of(true, v);
	  add_part_of_bc(cnode, parent[cnode], v, node_with_dfspos[min]);
	  u = v; //last_pnode(node_label_b[v], cnode);
	  if (!list_edges_upward_T0(parent[cnode], w))
	    error_handler(1, "-> Program terminated! (obstruction_edges_T0-02)");
	}
      obstr_edges_pnode(G, v, u);
    }

  obstr_edges_terminal(G, w, t2, node_with_dfspos[min]);

#ifdef _DEBUG_MODE_
  edge e;
  cout << "\n** List edges in obstruction: ";
  forall (e, obstruction_edges)
    cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
  cout << "\n";
  if (!CHECK_KURATOWSKI(G, obstruction_edges))
    error_handler(1, "-> Program terminated! (kuratowski_obstr_edges_T0!)");
#endif
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void sh_planar_impl::calc_info_3_terminals(
					   node &t1,
					   node &t2,
					   node &t3,
					   int &countmin,
					   int &countf,
					   node &cnode,
					   node &q)
{
  countmin = countf = 0;
  int min = label_b[t1];
  if (label_b[t2] < min)
    min = label_b[t2];
  if (label_b[t3] < min)
    min = label_b[t3];
  if (min == label_b[t1])
    countmin++;
  if (min == label_b[t2])
    countmin++;
  if (min == label_b[t3])
    countmin++;

  cnode = q = nil;
  node w1 = t1, w2 = t2, w3 = t3; 
  if (is_cnode(t1))
    w1 = parent(t1);
  if (is_cnode(t2))
    w2 = parent(t2);
  if (is_cnode(t3))
    w3 = parent(t3);
#ifdef _DEBUG_MODE_
  if (is_cnode(w1) || is_cnode(w2) || is_cnode(w3))
    error_handler(1, "-> Program terminated! (calc_info_3_terminals00)");
#endif
  node m1 = lca_between(w1, w2, parent),
    m2 = lca_between(w1, w3, parent),
    m3 = lca_between(w2, w3, parent);
  if (is_cnode(m1))
    m1 = active_cnode_of(true, m1);
  if (is_cnode(m2))
    m2 = active_cnode_of(true, m2);
  if (is_cnode(m3))
    m3 = active_cnode_of(true, m3);
  if (is_cnode(m1) && m1 == m2 && m2 == m3)
    {
      cnode = m1;
      node u1 = last_pnode(w1, cnode),
	u2 = last_pnode(w2, cnode),
	u3 = last_pnode(w3, cnode);
      if (u1 == t1)
	countf++;
      if (u2 == t2)
	countf++;
      if (u3 == t3)
	countf++;
    }
#ifdef _DEBUG_MODE_
  cout<<"countmin="<<countmin<<",countf="<<countf<<"\n";
#endif
  if (countf == 3)
    return;
  else
    cnode = nil;

  node q1 = lca_between(w1, w2, p0),
    q2 = lca_between(w1, w3, p0),
    q3 = lca_between(w2, w3, p0),
    k1 = t1, k2 = t2, k3 = t3;
  min = dfspos_num[q1];
  if (dfspos_num[q2] < min)
    {
      leda_swap(k2, k3);
      min = dfspos_num[q2];
    }
  if (dfspos_num[q3] < min)
    {
      k1 = t3;
      k2 = t2;
      k3 = t1;
      min = dfspos_num[q3];
    }
#ifdef _DEBUG_MODE_
  cout<<"w1="<<dfspos_num[w1]<<",w2="<<dfspos_num[w2]<<",w3="<<dfspos_num[w3]<<"\n";
  cout<<"q1="<<dfspos_num[q1]<<",q2="<<dfspos_num[q2]<<",q3="<<dfspos_num[q3]<<"\n";
  cout<<"min="<<min<<"\n";
  cout<<"k1="<<dfspos_num[k1]<<",k2="<<dfspos_num[k2]<<",k3="<<dfspos_num[k3]<<"\n";
#endif
  // is min a terminal node?
  if (min == dfspos_num[t1] || min == dfspos_num[t2] || min == dfspos_num[t3])
    {
      if (dfspos_num[k2] == min)
	leda_swap(k1, k2);
      if (dfspos_num[k3] == min) // q1 == q2 == q3;
	leda_swap(k1, k3);
#ifdef _DEBUG_MODE_
      cout<<"k1="<<dfspos_num[k1]<<",k2="<<dfspos_num[k2]<<",k3="<<dfspos_num[k3]<<"\n";
      if (dfspos_num[k1] != min)
	error_handler(1, "-> Program terminated! (calc_info_3_terminals02)");
#endif
      cnode = active_cnode_of(true, k1); // dfspos_num[k1] == min;

      int max = dfspos_num[q1];
      if (dfspos_num[q2] > max)
	max = dfspos_num[q2];
      if (dfspos_num[q3] > max)
	max = dfspos_num[q3];
      if (max >= dfspos_num[parent[cnode]])
	q = parent[cnode];
      else if (active_cnode_of(true, k2) != cnode)
	{
	  q = last_pnode(k2, cnode);
	  leda_swap(k2, k3);
	}
      else q = last_pnode(k3, cnode);
      t1 = k1;
      t2 = k2;
      t3 = k3;
#ifdef _DEBUG_MODE_
      cout<<"t1="<<dfspos_num[t1]<<",t2="<<dfspos_num[t2]<<",t3="<<dfspos_num[t3]<<"\n";
      if (cnode == nil)
	error_handler(1, "-> Program terminated! (calc_info_3_terminals01)");
      cout<<"cnode="<<dfspos_num[cnode]<<",q="<<dfspos_num[q]<<"\n";
      cout<<"t1="<<dfspos_num[t1]<<",t2="<<dfspos_num[t2]<<",t3="<<dfspos_num[t3]<<"\n";
#endif
      return;
    }
}
//-----------------------------------------------------------------------------------
