//-----------------------------------------------------------------------------------
// An implementation of Shih & Hsu's algorithm to test planarity.
//
// Given a graph G,
//
//   bool SH_PLANAR(graph_type& G, list<edge>& obstruction, bool embed)
//
// returns true if G is planar and false otherwise.
// If G is planar and embed == true and G is a map, then G is turned into 
// a plane map.
// If G is not planar and embed == true and G is a map, then obstr_edges 
// contains the edges of a subdivision of K_{3,3} or K_5 in G.
//
// Precondition:
// - G must be a bidirected graph;
// - to embed G, G must be loopfree and have no parallel edges.
//
// (Brief) Description:
//
// First, the program calculates a DFS (Depth First Search) tree T of G and
// numbers all nodes in G by their DFS post-order traversal number, from 1 to n,
// where n is the total number of nodes in G.
//
// All edges that are in G but not in T are called back-edges.
// If r is a node in T, T_r is the subtree of T rooted at r.
// An obstruction in a subgraph of G induces a subdivision of K_{3,3} or K_5 in G.
//
// The algorithm examines all nodes w of G from 1 to n.
// For each 2-connected component K formed by T_w and all back-edges
// from a node in T_w to w, the algorithm tests if K has an obstruction.
// If so, the program stops and returns false.
// If embed is true and G is a map, then the list of edges in the obstruction found
// is calculated.
// If is not detected any obstruction in K and embed is true and G is a map, an
// embedding of K is calculated and the program continues executing until all nodes
// of G are examined.
// If no obstruction is found in G and all nodes of G are examined, then the
// program returns true, and if embed is true and G is map, then G is turned into a
// plane map.
//
// Modules: sh_planar.h, sh_impl.h, sh_test.c, sh_aux.c, sh_embed.c, sh_obstr.c, 
//          dfspos.c, dfspos.h, bmlist.h.
//
//Credits:
//--------
//  
//  Alexandre Noma                <noma@ime.usp.br> -- supported by Fapesp.
//  Cristina Gomes Fernandes      <cris@ime.usp.br>
//  John M. Boyer                 <jboyer@acm.org>
//  Jose Coelho de Pina Junior    <coelho@ime.usp.br>
//
//  Institute of Mathematics and Statistics - University of Sao Paulo - Brazil
//-----------------------------------------------------------------------------------

#include "sh_impl.h"

#if defined(LEDA_NAMESPACE)
using namespace leda;
#endif

//-----------------------------------------------------------------------------------
// Tests if a graph G is planar.
// If G is planar and embed == true and G is a map, then G is turned into a plane map.
// If G is not planar and embed == true and G is a map, then obstr_edges contains the
// edges of a subdivision of K_{3,3} or K_5 in G (the construction of obstr_edges
// depends on the partial embedding of G).
// Returns true if G is planar, 
//         false otherwise.
//
// Precondition:
// - G is bidirected.
//-----------------------------------------------------------------------------------
bool SH_PLANAR(
	       graph_type& G,
               list<edge>& obstr_edges,
               bool to_embed)
{
  //#ifdef _DEBUG_MODE_
  if (!G.is_bidirected())
    error_handler(1, "-> Program terminated (sh_planar)! G must be bidirected!");
  //#endif

  obstr_edges.clear();

  bool planar, emb = false;
  if (to_embed && G.is_map())
    emb = true;

  // tests if G is planar;
  sh_planar_impl sh_test(G, emb, false);
  planar = sh_test.is_planar(G);
  if (!planar && emb)
    obstr_edges = sh_test.obstruction();

  // to turn G into a plane map, this algorithm requires that G is biconnected;
  if (planar && emb)
  { 
    list<edge> el;
    Make_Biconnected(G, el); 
    edge e;
#ifdef _DEBUG_MODE_
    if (!el.empty())
      cout<<"\n...Added "<<el.size()<<" edges to make G biconnected.\n";
#endif
    forall(e, el)
      { 
	edge x = G.new_edge(target(e), source(e));
	el.push(x);
	G.set_reversal(e, x);
      }
    // embeds G;
    sh_planar_impl sh_embed(G, emb, true);
    sh_embed.is_planar(G);
    // restores original graph;
    forall(e, el)
      G.del_edge(e);
  }

  return planar;
}
bool SH_PLANAR(graph_type& G)
{
  list<edge> el;
  return SH_PLANAR(G, el, false);
}

//-----------------------------------------------------------------------------------
// Tests if a graph G is planar.
// Returns true if it is, 
//         false otherwise.
// Precondition:
// - G is biconnected.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::is_planar(graph_type& G)
{
  int n = G.number_of_nodes();
  pre_processing(G);

#ifdef _DEBUG_MODE_
  edge e;
  forall_edges (e, G)
    if (T0_edge_in[target(e)] == e)
      G.hide_edge(e); 
  G.write("graph-debug.gw");
  G.restore_all_edges(); 
  if (embed && biconnected)
    cout << "** Embedding (traversed nodes):\n";
  else
    cout << "** Planarity Test (traversed nodes):\n";
#endif

  // list of terminal nodes for each component in T[V_G-L];
  node_array<list<node> > terminal_nodes(G);

  bool planar = true;
  for (int i = 1; i < n && planar; i++)
    {
      list<node> list_of_components; 
      node w = node_with_dfspos[i];

      // finds all terminal nodes and their respective components in T[V_G \ L];
     find_terminal_nodes(G, w, list_of_components, terminal_nodes);

      node comp;
      forall(comp, list_of_components)
        if (terminal_nodes[comp].size() > 0)
	  {
#ifdef _DEBUG_MODE_
	    cout << "       " << "*terminal nodes for w = " << dfspos_num[w] << ":\n";
	    node u;
	    cout << "                 in component (" << dfspos_num[comp] << "): ";
	    forall (u, terminal_nodes[comp])
	      {
		cout << dfspos_num[u];
		cout << ", ";
	      }
	    cout << "\n";
#endif
	    // creates a new c-node to represent current component;
	    node new_cnode = G.new_node();
	    dfspos_num[new_cnode] = -(++total_cnodes); // marks as c-node;

	    // does comp contain an obstruction?
	    if (find_obstruction(G, w, terminal_nodes[comp]))
	      {
		planar = false;
		break;
	      }
	    // else, calculates partial embedding of G;
	    if (embed)
	      calculate_partial_embedding(G, w, new_cnode, list_back_edges[comp],
					  terminal_nodes[comp]);

	    // calculates RBC, label_b, etc, for new c-node;
	    set_info_for_new_cnode(G, w, new_cnode, terminal_nodes[comp]);
	  }
    }

  // embeds root with all back-edges to root;
  if (planar && embed && biconnected)
    embed_root(G, n);

  // restores G;
  node u;
  forall_nodes(u, G)
    if (is_cnode(u)) 
      G.del_node(u);

#ifdef _DEBUG_MODE_
  if (!planar && embed)
    {
      cout << "\n** List edges in obstruction: ";
      forall (e, obstruction_edges)
	cout << "(" << dfspos_num[source(e)] << "," << dfspos_num[target(e)] << ") ";
      cout << "\n";
      if (!CHECK_KURATOWSKI(G, obstruction_edges))
	error_handler(1, "-> Program terminated! (kuratowski00!)");
    }
  if (planar && embed && biconnected)
    check_embedding(G);
#endif

  return planar;
}

//-----------------------------------------------------------------------------------
// Calculates:
// - DFS tree T_0 (initial T) of G represented as parent[u] for all nodes u in G;
// - dfspos_num[u] for all nodes u in T_0;
// - node_with_dfspos[i] = u iff dfspos_num[u] == i, for i = 1, .., n, where n is
//   the number of nodes in G;
// - largest_neighbor[u] for all nodes u in T_0;
// - label_b[u] for all nodes u in T_0 (see sh_impl.h for more details);
// - node_label_b[u] for all nodes u in T_0 (see sh_impl.h for more details);
// - children_in_T0[u] for each node u in T_0 (children of u in T_0 in decreasing
//   order by label_b -- needed by update_label_b);
//-----------------------------------------------------------------------------------
void sh_planar_impl::pre_processing(graph_type& G)
{
  int n = G.number_of_nodes();
  list<edge> E_T; // list of edges in T_0;
  E_T = DFS_POS(G, dfspos_num);

  edge e;
  forall(e, E_T)
    {
      node 
	u = source(e),
	v = target(e);
      parent[v] = u;
      T0_edge_in[v] = e; // needed by function is_back_edge;
    }
#ifdef _TESTING_MAPS_ONLY_
  forall(e, E_T)
    if (is_back_edge(e) || is_back_edge(G.reverse(e)))
      error_handler(1, "-> Program terminated! (preproc00)");
#endif 

  node u;
  forall_nodes(u, G)
    {
#ifdef _DEBUG_MODE_
      G[u] = dfspos_num[u];
#endif 
      node_with_dfspos[dfspos_num[u]] = u;
      largest_neighbor[u] = label_b[u] = dfspos_num[u];
      if (parent[u] != nil) // if u is not the root of T_0;
        largest_neighbor[u] = dfspos_num[parent[u]];
      if (embed)
        p0[u] = parent[u]; // saves T_0 for obstruction edges if G is not planar;
    }
  for (int i = 1; i <= n; i++)
    {
      node u = node_with_dfspos[i];
      forall_out_edges (e, u)  
	{
	  node v = target(e);
          if (label_b[u] < label_b[v])
	    label_b[u] = label_b[v];
	  if (largest_neighbor[u] < dfspos_num[v])
	    largest_neighbor[u] = dfspos_num[v];
	}
    }
#ifdef _DEBUG_MODE_
  cout << "** Labels (dfs[u], h[u], b[u]): ";
  for (int i = 1; i <= n; i++)
    {
      node u = node_with_dfspos[i];
      cout << "(" << dfspos_num[u] << "," << largest_neighbor[u] 
           << "," << label_b[u] << ") ";
    }
  cout << "\n";
#endif

  array<node> sorted_nodes(1, n);
  sort_nodes_increasing_order(G, label_b, sorted_nodes);
  for (int i = n; i >= 1; i--)
    {
      node
        u = sorted_nodes[i],
	v = parent[u];
      if (v != nil)
	children_in_T0[v].append(u);
    }
#ifdef _DEBUG_MODE_
  cout << "** List of children in T:\n";
  forall_nodes(u, G)
    {
      cout << "   " << dfspos_num[u] << ": ";
      node v;
      forall(v, children_in_T0[u])
	cout << dfspos_num[v] << "(" << label_b[v] << "), ";
      cout << "\n";
    }
#endif

  if (embed) // needed to calculate obstruction edges;
    for (int i = 1; i <= n; i++)
      {
	node u = node_with_dfspos[i];
	if (label_b[u] == largest_neighbor[u])
	  node_label_b[u] = u;
	else 
	  {
	    node v = children_in_T0[u].contents(children_in_T0[u].first());
	    node_label_b[u] = node_label_b[v];
	  }
      }
#ifdef _DEBUG_MODE_
  if (embed) // needed to calculate obstruction edges;
    forall_nodes(u, G)
      if (label_b[u] != largest_neighbor[node_label_b[u]])
	error_handler(1, "-> Program terminated! (preproc01)");
#endif
}

//-----------------------------------------------------------------------------------
// Calculates:
// - list_of_components: the list of all 2-connected components formed 
//   with T_w and all back-edges from a node in T_w to w 
//   (each component is represented by its subtree root in T);
// - terminal_nodes: the list of terminal nodes for each component found;
// - for each traversed c-node u, calculates counter[u]
//   (see sh_impl.h for more details).
// - if embed is true, also calculates:
//   - back_edge_out[u] for each neighbor u of w s.t. dfspos_num[u] < dfspos_num[w]
//     and u is neighbor of w via back-edge;
// Precondition:
// - state[u] == NOT_VISITED for all node u in G.
//-----------------------------------------------------------------------------------
void sh_planar_impl::find_terminal_nodes(
					 graph_type& G,
					 node w,
					 list<node>& list_of_components,
					 node_array<list<node> >& terminal_nodes)
{
  // to remove an element from list terminal_nodes in constant time;
  map<node, list_item> terminal_nodes_item;

  map<node, node> component_of;    
  list<node> traversed_nodes; // to restore state[u] for all traversed node u;
  list<edge> list_edges;

#ifdef _DEBUG_MODE_
  cout << "   " << dfspos_num[w] << ": ";
#endif
  component_of[w] = nil;
  state[w] = VISITED;
  traversed_nodes.append(w);
  edge e;
  forall_out_edges(e, w)  
    {
      node 
	u = target(e),
	neighbor_w = u;
      if (dfspos_num[u] < dfspos_num[w] && is_back_edge(e)) 
	{
          list<node> S; // helps calculate representant of 2-connected component;
	  if (embed)
	    {
	      back_edge_out[u] = G.reverse(e);
	      list_edges.append(G.reverse(e));
	    }
	  node terminal_node = nil;
          // looks for a terminal node, upward in T;
	  while (state[u] != VISITED && state[u] != TERMINAL) 
	    {
#ifdef _DEBUG_MODE_
	      cout << dfspos_num[u] << "(" << dfspos_num[parent[u]] << "), ";
#endif
	      if (is_cnode(parent[u]))
		{
		  node cnode = find_active_cnode(u, w, traversed_nodes);
		  counter[cnode]++;
		}
              // is u a candidate for a terminal node? 
	      if (terminal_node == nil && label_b[u] > dfspos_num[w])
		{
		  state[u] = TERMINAL;
		  terminal_node = u;
		}
	      else
		state[u] = VISITED;
	      traversed_nodes.append(u);
	      S.push(u);
	      u = parent[u];
	    } // end while

          // root of the component;
          if (u == w)
            {
              node c = S.pop();
	      component_of[c] = c;  
              list_of_components.append(c);
            }

          while (!S.empty())
            {
              node u = S.pop();
	      component_of[u] = component_of[parent[u]];
	    }

	  // did we find a terminal node?
	  if (terminal_node != nil)
	    {
	      node c = component_of[terminal_node];
	      // add the terminal node found in the terminal_nodes list 
              // of its respective component;
	      terminal_nodes_item[terminal_node] = 
                terminal_nodes[c].append(terminal_node);
	      neighbor_w_terminal[terminal_node] = neighbor_w;

	      // did we mark as terminal a non terminal node? 
	      if (state[u] == TERMINAL)
		{
		  state[u] = VISITED;
		  terminal_nodes[c].del(terminal_nodes_item[u]);
		  last_visited[terminal_node] = last_visited[u];
#ifdef _DEBUG_MODE_
		  cout << "changed " << dfspos_num[u];
		  cout << " to VISITED; ";
#endif
		}
	      else
		last_visited[terminal_node] = u;
	    }
	}
    }
#ifdef _DEBUG_MODE_
  cout << "\n";
#endif

  // groups all back-edges in T_w by their respective 2-connected component;
  forall(e, list_edges)
    list_back_edges[component_of[source(e)]].append(e);

  // restores state[u] for all traversed nodes u;
  node u;
  forall (u, traversed_nodes)
    state[u] = NOT_VISITED;
#ifdef _DEBUG_MODE_
  node v;
  forall(u, list_of_components)
    forall(v, terminal_nodes[u])
      if (is_cnode(v) && is_cnode(parent[v]))
	error_handler(1, "-> Program terminated! (findterminalnodes00)");
#endif
}

//-----------------------------------------------------------------------------------
// Checks if a 2-connected component K in T_v has an obstruction.
// Precondition:
// - all terminal nodes in K are in list terminal_nodes.
// Returns true if found an obstruction, false otherwise.
//-----------------------------------------------------------------------------------
bool sh_planar_impl::find_obstruction(graph_type& G, node w, list<node>& terminal_nodes)
{
  cnode_of_possible_k33_obstruction = nil; // reset global variable;

  list<node> list_terminal = terminal_nodes;

  switch(terminal_nodes.size())
    {
      case 1: 
	{
	  node t1 = list_terminal.pop();
	  if (test_obstruction_from_terminal_node(G, w, t1, w))
	    {
#ifdef _DEBUG_MODE_
              cout << "OBSTRUCTION_00\n";
#endif
	      return true;
	    }
	  if (cnode_of_possible_k33_obstruction != nil)
	    {
	      node v = find_node_with_label_b_greater_than_dfs_w(
		       false, G, w, cnode_of_possible_k33_obstruction);
	      if (v != nil)
		{
#ifdef _DEBUG_MODE_
		  cout<<"v="<<dfspos_num[v]<<",b="<<label_b[v]
		      <<",nl="<<dfspos_num[node_label_b[v]]<< "\n";
		  cout << "OBSTRUCTION_01\n";
#endif
		  if (embed)
		    obstruction_edges_possible_obstr_confirmed(G, w, t1, v);
		  return true;
		}
	    }
	}
	break;

      case 2: 
	{
	  node
	    t1 = list_terminal.pop(),
	    t2 = list_terminal.pop(),
	    m = lca_between_term_nodes(t1, t2);

	  node t12 = last_pnode(t1, m), t22 = last_pnode(t2, m);
	  if (test_obstruction_from_terminal_node(G, w, t1, t12))
	    {
#ifdef _DEBUG_MODE_
	      cout << "OBSTRUCTION_02\n";
#endif
	      return true;
	    }
	  if (cnode_of_possible_k33_obstruction != nil)
	    {
#ifdef _DEBUG_MODE_
	      cout << "OBSTRUCTION_03\n";
#endif
	      if (embed)
		{
		  node v = t2;
		  if (is_cnode(t2))
		    {
		      t2 = active_cnode_of(false, t2);
		      v = parent[t2];
		      node_label_b[v] = node_label_b[t2];
		      label_b[v] = label_b[t2];
		    }
		  obstruction_edges_possible_obstr_confirmed(G, w, t1, v);
		}
	      return true;
	    }

	  if (test_obstruction_from_terminal_node(G, w, t2, t22))
	    {
#ifdef _DEBUG_MODE_
	      cout << "OBSTRUCTION_04\n";
#endif
	      return true;
	    }
	  if (cnode_of_possible_k33_obstruction != nil)
	    {
#ifdef _DEBUG_MODE_
	      cout << "OBSTRUCTION_05\n";
#endif
	      if (embed)
		{
		  node v = t1;
		  if (is_cnode(t1))
		    {
		      t1 = active_cnode_of(false, t1);
		      v = parent[t1];
		      node_label_b[v] = node_label_b[t1];
		      label_b[v] = label_b[t1];
		    }
		  obstruction_edges_possible_obstr_confirmed(G, w, t2, v);
		}
	      return true;
	    }

	  node jl, jr;
	  if(is_cnode(m) && test_cnode_counter(G, m, w, t12, t22, jl, jr))
	    {
#ifdef _DEBUG_MODE_
	      cout << "OBSTRUCTION_06\n";
#endif
	      if (embed)
		obstruction_edges_cnode_counter(G, m, w, jl, jr, t1, t2);
	      return true;
	    }

	  // we still need to test if there is a node v in path Q_linha
	  // from parent[m] to a child of w, upward in T, has a descendant
	  // u which is not in Q_linha and label_b[u] > dfspos_num[w];
          node v = find_node_with_label_b_greater_than_dfs_w(true, G, w, m);
          if (v != nil) // v is p-node;
          {
#ifdef _DEBUG_MODE_
	    cout<<"v="<<dfspos_num[v]<<"("<<dfspos_num[parent[v]]<<"),b="<<label_b[v]
                <<",nl="<<dfspos_num[node_label_b[v]]<< "\n";
#endif
	    if (embed)
	      {
#ifdef _DEBUG_MODE_
		cout << "OBSTRUCTION_09\n";
#endif
		int countmin = 0;
		node vv = nil, mm = nil;
		if (is_cnode(m))
		  {
		    vv = lca_between(v, parent[m], p0);
		    if (vv == parent[m])
		      {
			label_b[parent[m]] = label_b[v];
			node_label_b[parent[m]] = node_label_b[v];
			v = parent[m];
		      }

		    int min = label_b[t1];
		    if (label_b[t2] < min)
		      min = label_b[t2];
		    if (label_b[v] < min)
		      min = label_b[v];
		    if (min == label_b[t1])
		      countmin++;
		    if (min == label_b[t2])
		      countmin++;
		    if (min == label_b[v])
		      countmin++;

		    node w1 = t1, w2 = t2;
		    if (is_cnode(t1))
		      w1 = parent[t1];
		    if (is_cnode(t2))
		      w2 = parent[t2];
		    mm = lca_between(w1, w2, p0);
		    if (mm == t1 || mm == t2)
		      {
			if (mm == t2)
			  {
			    leda_swap(t1, t2);  // makes t1 == t12;
			    leda_swap(t12, t22);
			  }
		      }
		    else
		      mm = nil;
#ifdef _DEBUG_MODE_
		    cout<<"t1="<<dfspos_num[t1]<<",t12="<<dfspos_num[t12]<<"\n";
		    cout<<"t2="<<dfspos_num[t2]<<",t22="<<dfspos_num[t22]<<"\n";
		    cout<<"countmin="<<countmin<<"\n";
		    if (mm != nil)
		      cout<<"mm="<<dfspos_num[mm]<<"\n";
#endif
		  }
		if (t1 == t12 && t2 == t22 && countmin >= 2 && vv == parent[m]) // m is c-node;
		  obstruction_edges_K5(G, w, m, t1, t2, nil);
		else if (mm != nil) // m is c-node;
		  {
		    if (t2 != t22 || vv != parent[m])
		      {
			node q = t22, k = parent[m];
			if (vv != parent[m])
			  {
			    q = parent[m];
			    k = v;
			  }
			obstruction_edges_countmin_2_3(G, w, m, t1, t2, nil, q, k);
		      }
		    else
		      obstruction_edges_countmin_1(G, w, m, t1, t2, nil);
		  }
		else
		  obstruction_edges_T0(G, w, t1, t2, nil, v);
		return true;
	      }
	  }
	}
	break;

      default: 
        {
#ifdef _DEBUG_MODE_
	  cout<<"OBSTRUCTION_08\n";
#endif
	  // if 3 or more terminal nodes, G has an obstruction;
	  if (embed)
	    {
	      node t1 = list_terminal.pop(),
		t2 = list_terminal.pop(),
		t3 = list_terminal.pop(),
		cnode, q;
	      int countmin, countf;
	      calc_info_3_terminals(t1, t2, t3, countmin, countf, cnode, q);
	      if (countf == 3 && countmin >= 2)
		obstruction_edges_K5(G, w, cnode, t1, t2, t3); // K5, bcycle;
	      else if (countf == 3)
		obstruction_edges_countmin_1(G, w, cnode, t1, t2, t3); // bcycle;
	      else if (cnode != nil)
		obstruction_edges_countmin_2_3(G, w, cnode, t1, t2, t3, q, nil); // bcycle;
	      else
		obstruction_edges_T0(G, w, t1, t2, t3, nil);    // T0;
	    }
	  return true;
	}
    }
  return false;
}

//-----------------------------------------------------------------------------------
// Calculates label_b, node_label_b, largest_neighbor and RBC for new_cnode,
// and inserts it in T.
// Initializes state and counter for new_cnode.
//-----------------------------------------------------------------------------------
void sh_planar_impl::set_info_for_new_cnode(
					      graph_type& G,
					      node w,
					      node new_cnode,
					      list<node>& terminal_nodes)
{
#ifdef _DEBUG_MODE_
  G[new_cnode] = dfspos_num[new_cnode];
#endif
  //largest_neighbor[new_cnode] = dfspos_num[w];
  label_b[new_cnode] = dfspos_num[w];
  if (embed) // needed to calculate obstruction edges;
    node_label_b[new_cnode] = nil;
  neighbor_w_terminal[new_cnode] = nil;
  parent[new_cnode] = w;
  state[new_cnode] = NOT_VISITED;
  counter[new_cnode] = 0;

  calculate_new_RBC(G, new_cnode, w, terminal_nodes);
  if (last_node_in_Q_linha != nil) // see find_node_with_label_b_greater_than_dfs_w;
    {
      parent[last_node_in_Q_linha] = new_cnode;
      last_node_in_Q_linha = nil; // resets global variable;
    }
  RBC[new_cnode].push(w);

  bmlist_item
    first = RBC[new_cnode].first_item(),
    last = RBC[new_cnode].last_item();
  activecnode[first] = new_cnode;
  activecnode[last] = new_cnode;

#ifdef _DEBUG_MODE_
  cout << "         RBC(" << dfspos_num[new_cnode] << ",b=" << label_b[new_cnode] << "): ";
  node u;
  bmlist_item it = RBC[new_cnode].first_item(), p = it;
  while (it != nil)
    //myforall(u, RBC[new_cnode])
    {
      u = RBC[new_cnode].contents(it);
      bmlist_item aux = it;
      it = RBC[new_cnode].next_item(it, p);
      p = aux;
      cout << "(" << dfspos_num[u] << ",b=" << label_b[u];
      node pu = parent[u];
      cout << "," << "p=" << dfspos_num[pu];
      pu = parent[pu];
      while (is_cnode(pu))
	{
	  cout << "," << dfspos_num[pu];
	  pu = parent[pu];
	}
      /*
      if (embed) // needed to calculate obstruction edges;
	cout << ":" << dfspos_num[node_label_b[u]] << "), ";
      cout << dfspos_num[u] << "(" << dfspos_num[RBC[new_cnode].contents(p)];
      if (it != nil)
        cout << "," << dfspos_num[RBC[new_cnode].contents(it)];
      */
      cout << "), ";
      if (p != ptritem[u] && p != RBC[new_cnode].first_item())
	{
	  cout << "ptrcontent="<<dfspos_num[RBC[new_cnode].contents(p)]
               <<",u="<<dfspos_num[u]<<"\n";
	  error_handler(1, "-> Program terminated! (set_info_for_new_cnode00)");
	}
    }
  cout << "\n";
#endif
}
//-----------------------------------------------------------------------------------
